from numbers import Real
from typing import Callable, Union, TypeVar

from fedot.core.composer.metrics import (ComputationTime, Accuracy, F1, Logloss, MAE,
                                         MAPE, SMAPE, MSE, MSLE, Metric, NodeNum, Precision, R2,
                                         RMSE, ROCAUC, Silhouette, StructuralComplexity)
from fedot.core.dag.graph import Graph
from fedot.core.utilities.data_structures import ComparableEnum as Enum


class MetricsEnum(Enum):
    pass


G = TypeVar('G', bound=Graph, covariant=True)
MetricCallable = Callable[[G], Real]
MetricType = Union[MetricCallable, MetricsEnum]


class QualityMetricsEnum(MetricsEnum):
    pass


class ComplexityMetricsEnum(MetricsEnum):
    node_num = 'node_number'
    structural = 'structural'
    computation_time = 'computation_time_in_seconds'


class ClusteringMetricsEnum(QualityMetricsEnum):
    silhouette = 'silhouette'


class ClassificationMetricsEnum(QualityMetricsEnum):
    ROCAUC = 'roc_auc'
    precision = 'precision'
    f1 = 'f1'
    logloss = 'neg_log_loss'
    ROCAUC_penalty = 'roc_auc_pen'
    accuracy = 'accuracy'


class RegressionMetricsEnum(QualityMetricsEnum):
    RMSE = 'rmse'
    MSE = 'mse'
    MSLE = 'neg_mean_squared_log_error'
    MAPE = 'mape'
    SMAPE = 'smape'
    MAE = 'mae'
    R2 = 'r2'
    RMSE_penalty = 'rmse_pen'


class MetricsRepository:
    _metrics_implementations = {
        # classification
        ClassificationMetricsEnum.ROCAUC: ROCAUC.get_value,
        ClassificationMetricsEnum.ROCAUC_penalty: ROCAUC.get_value_with_penalty,
        ClassificationMetricsEnum.f1: F1.get_value,
        ClassificationMetricsEnum.precision: Precision.get_value,
        ClassificationMetricsEnum.accuracy: Accuracy.get_value,
        ClassificationMetricsEnum.logloss: Logloss.get_value,

        # regression
        RegressionMetricsEnum.MAE: MAE.get_value,
        RegressionMetricsEnum.MSE: MSE.get_value,
        RegressionMetricsEnum.MSLE: MSLE.get_value,
        RegressionMetricsEnum.MAPE: MAPE.get_value,
        RegressionMetricsEnum.SMAPE: SMAPE.get_value,
        RegressionMetricsEnum.RMSE: RMSE.get_value,
        RegressionMetricsEnum.RMSE_penalty: RMSE.get_value_with_penalty,
        RegressionMetricsEnum.R2: R2.get_value,

        # clustering
        ClusteringMetricsEnum.silhouette: Silhouette.get_value,

        # structural
        ComplexityMetricsEnum.structural: StructuralComplexity.get_value,
        ComplexityMetricsEnum.node_num: NodeNum.get_value,
        ComplexityMetricsEnum.computation_time: ComputationTime.get_value
    }

    @staticmethod
    def metric_by_id(metric_id: MetricsEnum, default_callable: MetricCallable = None) -> MetricCallable:
        return MetricsRepository._metrics_implementations.get(metric_id, default_callable)

    @staticmethod
    def metric_class_by_id(metric_id: MetricsEnum) -> Metric:
        return MetricsRepository._metrics_implementations[metric_id].__self__()



import traceback
from typing import List, Optional, Union

from fedot.api.api_utils.assumptions.assumptions_builder import AssumptionsBuilder
from fedot.api.api_utils.presets import change_preset_based_on_initial_fit
from fedot.api.time import ApiTime
from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.log import default_log
from fedot.core.pipelines.pipeline import Pipeline


class AssumptionsHandler:
    def __init__(self, data: InputData):
        """
        Class for handling operations related with assumptions

        :param data: data for pipelines
        """
        self.log = default_log(self)
        self.data = data

    def propose_assumptions(self,
                            initial_assumption: Union[List[Pipeline], Pipeline, None],
                            available_operations: List) -> List[Pipeline]:
        """
        Method to propose initial assumptions if needed

        :param initial_assumption: initial assumption given by user
        :param available_operations: list of available operations defined by user
        """

        if initial_assumption is None:
            assumptions_builder = AssumptionsBuilder \
                .get(self.data) \
                .from_operations(available_operations)
            initial_assumption = assumptions_builder.build()
        elif isinstance(initial_assumption, Pipeline):
            initial_assumption = [initial_assumption]
        return initial_assumption

    def fit_assumption_and_check_correctness(self,
                                             pipeline: Pipeline,
                                             pipelines_cache: Optional[OperationsCache] = None,
                                             preprocessing_cache: Optional[PreprocessingCache] = None) -> Pipeline:
        """
        Check if initial pipeline can be fitted on a presented data

        :param pipeline: pipeline for checking
        :param pipelines_cache: Cache manager for fitted models, optional.
        :param preprocessing_cache: Cache manager for optional preprocessing encoders and imputers, optional.
        """
        try:
            data_train, data_test = train_test_data_setup(self.data)
            self.log.info('Initial pipeline fitting started')
            # load preprocessing
            pipeline.try_load_from_cache(pipelines_cache, preprocessing_cache)
            pipeline.fit(data_train)

            if pipelines_cache is not None:
                pipelines_cache.save_pipeline(pipeline)
            if preprocessing_cache is not None:
                preprocessing_cache.add_preprocessor(pipeline)

            pipeline.predict(data_test)
            self.log.info('Initial pipeline was fitted successfully')

        except Exception as ex:
            self._raise_evaluating_exception(ex)
        return pipeline

    def _raise_evaluating_exception(self, ex: Exception):
        fit_failed_info = f'Initial pipeline fit was failed due to: {ex}.'
        advice_info = f'{fit_failed_info} Check pipeline structure and the correctness of the data'
        self.log.info(fit_failed_info)
        print(traceback.format_exc())
        raise ValueError(advice_info)

    def propose_preset(self, preset: Union[str, None], timer: ApiTime, n_jobs: int) -> str:
        """
        Proposes the most suitable preset for current data

        :param preset: predefined preset
        :param timer: ApiTime object
        :param n_jobs: n_jobs parameter

        """
        if not preset or preset == 'auto':
            preset = change_preset_based_on_initial_fit(timer, n_jobs)
            self.log.info(f"Preset was changed to {preset}")
        return preset



import os

import pandas as pd

from examples.simple.classification.classification_pipelines import classification_complex_pipeline
from fedot.core.data.data import InputData
from fedot.core.utils import fedot_project_root
from fedot.explainability.explainers import explain_pipeline


def run_pipeline_explain():
    # Specifying paths
    train_data_path = os.path.join(fedot_project_root(), 'cases', 'data', 'cancer', 'cancer_train.csv')
    figure_path = 'pipeline_explain_example.png'

    # Feature and class names for visualization
    feature_names = pd.read_csv(train_data_path, index_col=0, nrows=0).columns.tolist()
    target_name = feature_names.pop()
    target = pd.read_csv(train_data_path, usecols=[target_name])[target_name]
    class_names = target.unique().astype(str).tolist()

    # Data load
    train_data = InputData.from_csv(train_data_path)

    # Pipeline composition
    pipeline = classification_complex_pipeline()

    # Pipeline fitting
    pipeline.fit(train_data)

    # Pipeline explaining
    explainer = explain_pipeline(pipeline, data=train_data, method='surrogate_dt', visualization=True)

    # Visualizing explanation and saving the plot
    print(f'Built surrogate model: {explainer.surrogate_str}')
    explainer.visualize(save_path=figure_path, dpi=200, feature_names=feature_names, class_names=class_names,
                        precision=6)


if __name__ == '__main__':
    run_pipeline_explain()



from typing import Optional, Sequence

from fedot.core.dag.verification_rules import DEFAULT_DAG_RULES
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.pipelines.pipeline_advisor import PipelineChangeAdvisor
from fedot.core.dag.graph_verifier import VerifierRuleType
from fedot.core.pipelines.adapters import PipelineAdapter
from fedot.core.optimisers.optimizer import GraphGenerationParams
from fedot.core.pipelines.pipeline_node_factory import PipelineOptNodeFactory
from fedot.core.repository.operation_types_repository import get_operations_for_task
from fedot.core.repository.tasks import Task


def get_pipeline_generation_params(requirements: Optional[PipelineComposerRequirements] = None,
                                   rules_for_constraint: Sequence[VerifierRuleType] = tuple(DEFAULT_DAG_RULES),
                                   task: Optional[Task] = None) -> GraphGenerationParams:
    if requirements is None and task is not None:
        ops = get_operations_for_task(task)
        requirements = PipelineComposerRequirements(primary=ops, secondary=ops)
    else:
        requirements = requirements or PipelineComposerRequirements()
    advisor = PipelineChangeAdvisor(task)
    node_factory = PipelineOptNodeFactory(requirements, advisor) if requirements else None
    graph_generation_params = GraphGenerationParams(adapter=PipelineAdapter(),
                                                    rules_for_constraint=rules_for_constraint,
                                                    advisor=advisor,
                                                    node_factory=node_factory)
    return graph_generation_params



from collections import namedtuple
from os import makedirs
from os.path import exists, join
from typing import List, Optional, Type

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.quality_metrics_repository import MetricsRepository
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.node_sa_approaches import NodeAnalyzeApproach, NodeDeletionAnalyze
from fedot.sensitivity.nodes_sensitivity import NodesAnalysis
from fedot.utilities.define_metric_by_task import MetricByTask

MTAMetaParams = namedtuple('MTAMetaParams', ['delta', 'worst_node_score'])


class MultiTimesAnalyze:
    """
    Multi-Times-Analyze approach is used for Pipeline size decrease
    using node-deletion-algorithm defined in MultiTimesAnalyze.analyze method

    :param pipeline: pipeline object to analyze
    :param train_data: data used for Pipeline training
    :param test_data: data used for getting prediction
    :param valid_data: used for modification validation
    :param case_name: used for uniq result directory name
    :param approaches: methods applied to nodes to modify
        the pipeline or analyze certain operations.\
    Defaults: NodeDeletionAnalyze.
    """

    default_mta_meta_params = MTAMetaParams(10e-3, 1.1)

    def __init__(self, pipeline: Pipeline, train_data: InputData,
                 test_data: InputData, valid_data: InputData,
                 case_name: str, path_to_save: str = None,
                 approaches: Optional[List[Type[NodeAnalyzeApproach]]] = None):
        self.pipeline = pipeline
        self.original_pipeline_len = self.pipeline.length
        self.train_data = train_data
        self.test_data = test_data
        self.valid_data = valid_data
        self.case_name = case_name
        self.path_to_save = \
            join(default_fedot_data_dir(),
                 'sensitivity', 'mta_analysis', f'{case_name}') \
                if path_to_save is None else path_to_save
        self.approaches = [NodeDeletionAnalyze] if approaches is None else approaches
        self.log = default_log(self)

    def analyze(self, visualization=False, meta_params: MTAMetaParams = None) -> float:
        """
        Algorithm:
        1. Analyze pipeline
        2. Defines potential 'bad' nodes within iteration
        3. Choose the worst one and delete it
        3. Repeat 1-3 till the condition: no more 'bad' nodes(worst_node_score<=1) or len(Pipeline) < 2

        :param meta_params: limiting params for sensitivity index:
         MTAMetaParams(delta, worst_node_score)
         (defaults: 10e-3, 1.1 correspondingly).
        :param visualization: boolean flag for pipeline structure visualization. Default: False

        :return ratio of number of deleted nodes to overall Pipeline length
        """

        if not meta_params:
            meta_params = self.default_mta_meta_params

        total_nodes_deleted = 0
        iteration_index = 1
        worst_node_score = meta_params.worst_node_score
        while worst_node_score > 1.0 + meta_params.delta and self.pipeline.length > 2:
            self.log.info('new iteration of MTA deletion analysis')
            iteration_result_path = join(self.path_to_save, f'iter_{iteration_index}')
            pipeline_analysis_result = self._pipeline_analysis(result_path=iteration_result_path,
                                                               visualization=visualization)

            deletion_scores = [node['NodeDeletionAnalyze'] for node in pipeline_analysis_result.values()]
            worst_node_score = max(deletion_scores)

            if worst_node_score > 1.0 + meta_params.delta:
                worst_node_index = deletion_scores.index(worst_node_score)
                self.pipeline.delete_node(self.pipeline.nodes[worst_node_index])
                total_nodes_deleted += 1

            iteration_index += 1

        self.log.info('finish MTA')
        return self._length_reduction_ratio(total_nodes_deleted)

    def _length_reduction_ratio(self, number_of_deleted_nodes: int):
        return number_of_deleted_nodes / self.original_pipeline_len

    def _pipeline_analysis(self, result_path, visualization=False):
        if not exists(result_path):
            makedirs(result_path)

        if visualization:
            self._visualize(name=self.case_name, path=result_path)

        self.pipeline.fit_from_scratch(self.train_data)

        self.log.info('Start Pipeline Analysis')

        pipeline_analysis_result = NodesAnalysis(pipeline=self.pipeline, train_data=self.train_data,
                                                 test_data=self.test_data,
                                                 path_to_save=result_path,
                                                 approaches=self.approaches).analyze()
        self.log.info("End Pipeline Analysis")

        return pipeline_analysis_result

    def _visualize(self, name, path):
        image_path = join(path, f'{name}.png')
        self.pipeline.show(save_path=image_path)

    def get_metric(self):
        self.pipeline.fit(self.train_data)
        predicted = self.pipeline.predict(self.valid_data)
        metric_value = MetricByTask.compute_default_metric(
            self.valid_data.task.task_type,
            true=self.valid_data,
            predicted=predicted
        )

        return metric_value



import os
import warnings

import numpy as np
import pandas as pd

from examples.advanced.time_series_forecasting.composing_pipelines import get_border_line_info
from examples.simple.time_series_forecasting.ts_pipelines import *
from examples.simple.time_series_forecasting.tuning_pipelines import visualise
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.pipeline import Pipeline

from fedot.core.pipelines.ts_wrappers import out_of_sample_ts_forecast
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum, Task, TsForecastingParams

from fedot.core.utils import fedot_project_root

warnings.filterwarnings('ignore')
np.random.seed(2020)

datasets = {
    'australia': f'{fedot_project_root()}/examples/data/ts/australia.csv',
    'beer': f'{fedot_project_root()}/examples/data/ts/beer.csv',
    'salaries': f'{fedot_project_root()}/examples/data/ts/salaries.csv',
    'stackoverflow': f'{fedot_project_root()}/examples/data/ts/stackoverflow.csv',
    'test_sea': os.path.join(fedot_project_root(), 'test', 'data', 'simple_sea_level.csv')}


def run_multistep(dataset: str, pipeline: Pipeline, step_forecast: int = 10, future_steps: int = 5,
                  visualisation=False):
    """ Example of out-of-sample ts forecasting using custom pipelines
    :param dataset: name of dataset
    :param pipeline: pipeline to use
    :param step_forecast: horizon to train model. Real horizon = step_forecast * future_steps
    :param future_steps: number of future steps
    :param visualisation: is visualisation needed
    """
    # show initial pipeline
    pipeline.print_structure()

    horizon = step_forecast * future_steps
    time_series = pd.read_csv(datasets[dataset])
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=step_forecast))

    idx = np.arange(len(time_series['idx'].values))
    time_series = time_series['Level'].values
    train_input = InputData(idx=idx,
                            features=time_series,
                            target=time_series,
                            task=task,
                            data_type=DataTypesEnum.ts)
    train_data, test_data = train_test_data_setup(train_input)

    pipeline.fit(train_data)

    predict = out_of_sample_ts_forecast(pipeline=pipeline,
                                        input_data=test_data,
                                        horizon=horizon)

    plot_info = [{'idx': idx,
                  'series': time_series,
                  'label': 'Actual time series'},
                 {'idx': np.arange(test_data.idx[0], test_data.idx[0] + predict.shape[0]),
                  'series': predict,
                  'label': 'Forecast'},
                 get_border_line_info(np.arange(test_data.idx[0] + 1)[-1], predict, time_series, 'train|test'),
                 get_border_line_info(np.arange(test_data.idx[-1] + 1)[-1], predict, time_series, 'End of test',
                                      'gray')]

    # plot lines
    if visualisation:
        visualise(plot_info)


if __name__ == '__main__':
    run_multistep("australia", ts_ar_pipeline(), step_forecast=10, visualisation=True)



from copy import deepcopy
from datetime import timedelta
from typing import Callable, Union, Optional

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.operation import Operation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.operation_types_repository import OperationMetaInfo, \
    atomized_model_type
from fedot.core.utils import make_pipeline_generator
from fedot.preprocessing.preprocessing import DataPreprocessor


class AtomizedModel(Operation):
    """ Class which replace Operation class for AtomizedModel object """

    def __init__(self, pipeline: 'Pipeline'):
        if not pipeline.root_node:
            raise ValueError('AtomizedModel could not create instance of empty Pipeline.')

        super().__init__(operation_type=atomized_model_type())
        self.pipeline = pipeline
        self.unique_id = self.pipeline.root_node.descriptive_id
        self.atomized_preprocessor = DataPreprocessor()

    def fit(self, params: Optional[Union[OperationParameters, dict]], data: InputData,
            use_cache: bool = True):

        copied_input_data = deepcopy(data)
        copied_input_data = self.atomized_preprocessor.obligatory_prepare_for_fit(copied_input_data)

        predicted_train = self.pipeline.fit(input_data=copied_input_data)
        fitted_atomized_operation = self.pipeline

        return fitted_atomized_operation, predicted_train

    def predict(self, fitted_operation, data: InputData,
                params: Optional[Union[OperationParameters, dict]] = None, output_mode: str = 'default'):

        # Preprocessing applied
        copied_input_data = deepcopy(data)
        copied_input_data = self.atomized_preprocessor.obligatory_prepare_for_predict(copied_input_data)

        prediction = fitted_operation.predict(input_data=copied_input_data, output_mode=output_mode)
        prediction = self.assign_tabular_column_types(prediction, output_mode)
        return prediction

    def predict_for_fit(self, fitted_operation, data: InputData, params: Optional[OperationParameters] = None,
                        output_mode: str = 'default'):
        return self.predict(fitted_operation, data, params, output_mode)

    def fine_tune(self, metric_function: Callable,
                  input_data: InputData = None, iterations: int = 50,
                  timeout: int = 5):
        """ Method for tuning hyperparameters """
        tuner = TunerBuilder(input_data.task)\
            .with_tuner(PipelineTuner)\
            .with_metric(metric_function)\
            .with_iterations(iterations)\
            .with_timeout(timedelta(minutes=timeout))\
            .build(input_data)
        tuned_pipeline = tuner.tune(self.pipeline)
        tuned_atomized_model = AtomizedModel(tuned_pipeline)
        return tuned_atomized_model

    @property
    def metadata(self) -> OperationMetaInfo:
        generator = make_pipeline_generator(self.pipeline)
        tags = set()

        for node in generator:
            tags.update(node.operation_tags)

        root_node = self.pipeline.root_node
        supported_strategies = None
        allowed_positions = ['any']
        tags = list(tags)

        operation_info = OperationMetaInfo(root_node.operation.supplementary_data.id,
                                           root_node.operation.supplementary_data.input_types,
                                           root_node.operation.supplementary_data.output_types,
                                           root_node.operation.supplementary_data.task_type,
                                           supported_strategies, allowed_positions,
                                           tags)
        return operation_info

    def description(self, operation_params: Optional[dict]):
        operation_type = self.operation_type
        operation_length = self.pipeline.length
        operation_depth = self.pipeline.depth
        operation_id = self.unique_id
        operation_types = {}

        for node in self.pipeline.nodes:
            if node.operation.operation_type in operation_types:
                operation_types[node.operation.operation_type] += 1
            else:
                operation_types[node.operation.operation_type] = 1

        return f'{operation_type}_length:{operation_length}_depth:{operation_depth}' \
               f'_types:{operation_types}_id:{operation_id}'

    @staticmethod
    def assign_tabular_column_types(output_data: OutputData,
                                    output_mode: str) -> OutputData:
        """ There is no need to perform any column types determination for nested pipelines """
        return output_data



from typing import Optional

from fedot.core.operations.model import Model
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_operations_for_task, \
    atomized_model_type
from fedot.core.repository.tasks import Task, TaskTypesEnum

ERROR_PREFIX = 'Invalid pipeline configuration:'


def has_correct_operations_for_task(pipeline: Pipeline, task_type: Optional[TaskTypesEnum] = None):
    if task_type and not task_type in pipeline.root_node.operation.acceptable_task_types:
        raise ValueError(f'{ERROR_PREFIX} Pipeline has incorrect operations positions')
    return True


def has_primary_nodes(pipeline: Pipeline):
    if not any(node for node in pipeline.nodes if isinstance(node, PrimaryNode)):
        raise ValueError(f'{ERROR_PREFIX} Pipeline does not have primary nodes')
    return True


def has_final_operation_as_model(pipeline: Pipeline):
    """ Check if the operation in root node is model or not """
    root_node = pipeline.root_node
    # TODO @YamLyubov refactor check for AtomizedModel (fix circular import)
    if type(root_node.operation) is not Model and root_node.operation.operation_type != atomized_model_type():
        raise ValueError(f'{ERROR_PREFIX} Root operation is not a model')

    return True


def has_no_conflicts_with_data_flow(pipeline: Pipeline):
    """ Check if the pipeline contains incorrect connections between nodes """
    operation_repo = OperationTypesRepository(operation_type='data_operation')
    forbidden_parents_combination = operation_repo.suitable_operation()
    forbidden_parents_combination = set(forbidden_parents_combination)

    for node in pipeline.nodes:
        parent_nodes = node.nodes_from

        if parent_nodes is not None and len(parent_nodes) > 1:
            # There are several parents
            operation_names = []
            for parent in parent_nodes:
                operation_names.append(parent.operation.operation_type)

            # If operations are identical
            if len(set(operation_names)) == 1:
                # And if it is forbidden to combine them
                if operation_names[0] in forbidden_parents_combination:
                    raise ValueError(f'{ERROR_PREFIX} Pipeline has incorrect subgraph with identical data operations')
    return True


def has_correct_data_connections(pipeline: Pipeline):
    """ Check if the pipeline contains incorrect connections between operation for different data types """
    _repo = OperationTypesRepository(operation_type='all')

    for node in pipeline.nodes:
        parent_nodes = node.nodes_from

        if parent_nodes is not None and len(parent_nodes) > 0:
            for parent_node in parent_nodes:
                if 'custom' in str(parent_node) or 'custom' in str(node):
                    return True

                current_nodes_supported_data_types = _repo.operation_info_by_id(node.operation.operation_type)
                parent_node_supported_data_types = _repo.operation_info_by_id(parent_node.operation.operation_type)

                if current_nodes_supported_data_types is None:
                    # case for atomic model
                    return True

                node_dtypes = set(current_nodes_supported_data_types.input_types)
                parent_dtypes = set(parent_node_supported_data_types.output_types) \
                    if parent_node_supported_data_types else node_dtypes
                if len(set.intersection(node_dtypes, parent_dtypes)) == 0:
                    raise ValueError(f'{ERROR_PREFIX} Pipeline has incorrect data connections')

    return True


def is_pipeline_contains_ts_operations(pipeline: Pipeline):
    """ Function checks is the model contains operations for time series
    forecasting """

    # Get time series specific operations with tag "non_lagged"
    ts_operations = get_operations_for_task(task=Task(TaskTypesEnum.ts_forecasting),
                                            tags=["non_lagged"], mode='all')

    # List with operations in considering pipeline
    operations_in_pipeline = []
    for node in pipeline.nodes:
        operations_in_pipeline.append(node.operation.operation_type)

    if len(set(ts_operations) & set(operations_in_pipeline)) > 0:
        return True
    else:
        raise ValueError(f'{ERROR_PREFIX} pipeline not contains operations for time series processing')


def has_no_data_flow_conflicts_in_ts_pipeline(pipeline: Pipeline):
    """ Function checks the correctness of connection between nodes """

    task = Task(TaskTypesEnum.ts_forecasting)
    models = get_operations_for_task(task=task, mode='model')
    # Preprocessing not only for time series
    non_ts_data_operations = get_operations_for_task(task=task,
                                                     mode='data_operation',
                                                     tags=["non_applicable_for_ts"])
    ts_data_operations = get_operations_for_task(task=task,
                                                 mode='data_operation',
                                                 tags=["non_lagged"])
    # Remove lagged and sparse lagged transformation
    ts_data_operations.remove('lagged')
    ts_data_operations.remove('sparse_lagged')
    ts_data_operations.remove('exog_ts')

    # Dictionary as {'current operation in the node': 'parent operations list'}
    # TODO refactor
    wrong_connections = {'lagged': models + non_ts_data_operations + ['lagged', 'sparse_lagged', 'exog_ts'],
                         'sparse_lagged': models + non_ts_data_operations + ['lagged', 'sparse_lagged', 'exog_ts'],
                         'ar': models + non_ts_data_operations + ['lagged', 'sparse_lagged', 'exog_ts'],
                         'arima': models + non_ts_data_operations + ['lagged', 'sparse_lagged', 'exog_ts'],
                         'ridge': ts_data_operations, 'linear': ts_data_operations,
                         'lasso': ts_data_operations, 'dtreg': ts_data_operations,
                         'knnreg': ts_data_operations, 'scaling': ts_data_operations,
                         'xgbreg': ts_data_operations, 'adareg': ts_data_operations,
                         'gbr': ts_data_operations, 'treg': ts_data_operations,
                         'rfr': ts_data_operations, 'svr': ts_data_operations,
                         'sgdr': ts_data_operations, 'normalization': ts_data_operations,
                         'kernel_pca': ts_data_operations, 'poly_features': ts_data_operations,
                         'ransac_lin_reg': ts_data_operations, 'ransac_non_lin_reg': ts_data_operations,
                         'rfe_lin_reg': ts_data_operations, 'rfe_non_lin_reg': ts_data_operations,
                         'pca': ts_data_operations,
                         'gaussian_filter': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'diff_filter': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'smoothing': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'cut': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'ts_naive_average': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'locf': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'ets': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'polyfit': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'clstm': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'glm': ['lagged', 'sparse_lagged', 'exog_ts'],
                         'stl_arima': ['lagged', 'sparse_lagged', 'exog_ts'],
                         }

    for node in pipeline.nodes:
        # Operation name in the current node
        current_operation = node.operation.operation_type
        parent_nodes = node.nodes_from

        if parent_nodes is not None:
            # There are several parents for current node or at least 1
            for parent in parent_nodes:
                parent_operation = parent.operation.operation_type

                forbidden_parents = wrong_connections.get(current_operation)
                if forbidden_parents is not None:
                    __check_connection(parent_operation, forbidden_parents)
    return True


def only_non_lagged_operations_are_primary(pipeline: Pipeline):
    """ Only time series specific operations could be placed in primary nodes """

    # Check only primary nodes
    for node in pipeline.nodes:
        if type(node) == PrimaryNode and DataTypesEnum.ts not in node.operation.metadata.input_types:
            raise ValueError(
                f'{ERROR_PREFIX} Pipeline for forecasting has not non_lagged preprocessing in primary nodes')

    return True


def has_no_conflicts_in_decompose(pipeline: Pipeline):
    """ The function checks whether the 'class_decompose' or 'decompose' operation has two ancestors """

    for decomposer in ['decompose', 'class_decompose']:
        decompose_nodes = pipeline.get_nodes_by_name(decomposer)
        if len(decompose_nodes) != 0:
            # Launch check decomposers
            __check_decomposer_has_two_parents(nodes_to_check=decompose_nodes)
            __check_decompose_parent_position(nodes_to_check=decompose_nodes)

    return True


def has_correct_data_sources(pipeline: Pipeline):
    """ Checks that data sources and other nodes are not mixed """

    is_data_source_in_names_conds = ['data_source' in str(n) for n in pipeline.nodes if isinstance(n, PrimaryNode)]

    if any(is_data_source_in_names_conds) and not all(is_data_source_in_names_conds):
        raise ValueError(f'{ERROR_PREFIX} Data sources are mixed with other primary nodes')
    return True


def has_parent_contain_single_resample(pipeline: Pipeline):
    """ 'Resample' should be single parent node for child operation. """

    for node in pipeline.nodes:
        if node.operation.operation_type == 'resample':
            children_nodes = pipeline.node_children(node)
            for child_node in children_nodes:
                if len(child_node.nodes_from) > 1:
                    raise ValueError(f'{ERROR_PREFIX} Resample node is not single parent node for child operation')

    return True


def has_no_conflicts_during_multitask(pipeline: Pipeline):
    """
    Now if the classification task is solved, one part of the pipeline can solve
    the regression task if used after class_decompose. If class_decompose is followed
    by a classification operation, then this pipelining is incorrect.
    Validation perform only for classification pipelines.
    """

    classification_operations = get_operations_for_task(task=Task(TaskTypesEnum.classification), mode='all')
    pipeline_operations = [node.operation.operation_type for node in pipeline.nodes]
    pipeline_operations = set(pipeline_operations)

    number_of_unique_pipeline_operations = len(pipeline_operations)
    pipeline_operations_for_classification = set(classification_operations).intersection(pipeline_operations)

    if len(pipeline_operations_for_classification) == 0:
        return True

    if 'class_decompose' not in pipeline_operations:
        # There are no decompose operations in the pipeline
        if number_of_unique_pipeline_operations != len(pipeline_operations_for_classification):
            # There are operations in the pipeline that solve different tasks
            __check_multitask_operation_location(pipeline, classification_operations)

    return True


def has_no_conflicts_after_class_decompose(pipeline: Pipeline):
    """
    After the class_decompose operation, a regression model is required.
    Validation perform only for classification pipelines.
    """
    error_message = f'{ERROR_PREFIX} After classification decompose it is required to use regression model'
    pipeline_operations = [node.operation.operation_type for node in pipeline.nodes]
    if 'class_decompose' not in pipeline_operations:
        return True

    regression_operations = get_operations_for_task(task=Task(TaskTypesEnum.regression), mode='all')

    # Check for correct descendants after classification decompose
    for node in pipeline.nodes:
        parent_operations = [node.operation.operation_type for node in node.nodes_from]
        if 'class_decompose' in parent_operations:
            # Check is this model for regression task
            if node.operation.operation_type not in regression_operations:
                raise ValueError(error_message)

    return True


def __check_connection(parent_operation, forbidden_parents):
    if parent_operation in forbidden_parents:
        raise ValueError(f'{ERROR_PREFIX} Pipeline has incorrect subgraph with wrong parent nodes combination')


def __check_decompose_parent_position(nodes_to_check: list):
    """ Function check if the data flow before decompose operation is correct
    or not

    :param nodes_to_check: list with decompose nodes in the pipeline
    """
    for decompose_node in nodes_to_check:
        parents = decompose_node.nodes_from
        model_parent = parents[0]

        if type(model_parent.operation) is not Model:
            raise ValueError(f'{ERROR_PREFIX} For decompose operation Model as first parent is required')


def __check_decomposer_has_two_parents(nodes_to_check: list):
    """ Function check if there are two parent nodes for decompose operation

    :param nodes_to_check: list with decompose nodes in the pipeline
    """

    for decompose_node in nodes_to_check:
        parents = decompose_node.nodes_from

        if parents is None:
            raise ValueError('Decompose operation has no parents')
        elif len(parents) != 2:
            raise ValueError(f'{ERROR_PREFIX} Two parents for decompose node were'
                             f' expected, but {len(parents)} were given')


def __check_multitask_operation_location(pipeline: Pipeline, operations_for_classification: list):
    """
    Investigate paths for different tasks in the pipeline. If the pipeline solves
    several tasks simultaneously and there are no transitive operations in its
    structure (e.g. class_decompose), then the side branches must start from the
    primary node (nodes)
    """
    # TODO refactor to implement check via PipelineStructureExplorer
    primary_operations = []
    for node in pipeline.nodes:
        if isinstance(node, PrimaryNode):
            primary_operations.append(node.operation.operation_type)

    primary_operations = set(primary_operations)
    unique_primary_operations_number = len(primary_operations)

    primary_operations_for_classification = set(operations_for_classification).intersection(primary_operations)

    if unique_primary_operations_number != len(primary_operations_for_classification):
        # There are difference in tasks are in the primary nodes
        return True
    else:
        raise ValueError(f'{ERROR_PREFIX} Current pipeline can not solve multitask problem')



import os
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, Optional, Union

import joblib
import numpy as np

from fedot.core.log import default_log
from fedot.core.pipelines.node import Node


class OperationTemplateAbstract(ABC):
    """
    Base class used for create different types of operation ("atomized_operation"
    or others like ("knn", "rf")).
    Atomized_operation is pipeline which can be used like general operation.
    """

    def __init__(self):
        self.operation_id = None
        self.operation_type = None
        self.nodes_from = None

        self.log = default_log(self)

    @abstractmethod
    def _operation_to_template(self, node: Node, operation_id: int, nodes_from: list):
        """
        Preprocessing for local fields
        :param node: current node
        :param operation_id: operation id in pipeline
        :param nodes_from: parents operation's id
        """

    @abstractmethod
    def import_json(self, operation_object: dict):
        """
        Parse JSON like object and fill local fields
        :param operation_object: JSON like object to parse
        """

    @abstractmethod
    def convert_to_dict(self) -> dict:
        """
        Transform all object's parameters to dictionary.
        :params path: string path to save.
        :return dict: dictionary with object parameters.
        """

    def _validate_json_operation_template(self, operation_object: dict, required_fields: list):
        """
        Check whether there are fields in the dictionary.
        :params operation_object: dictionary to check
        :params required_fields: list of fields name
        """

        for field in required_fields:
            if field not in operation_object:
                message = f"Required field '{field}' is expected, but not found."
                self.log.error(message)
                raise RuntimeError(message)


class OperationTemplate(OperationTemplateAbstract):
    def __init__(self, node: Node = None, operation_id: int = None,
                 nodes_from: list = None):
        super().__init__()
        self.operation_name = None
        self.custom_params = None
        self.params = None
        self.fitted_operation = None
        self.fitted_operation_path = None
        self.rating = None

        if node:
            self._operation_to_template(node, operation_id, nodes_from)

    def _operation_to_template(self, node: Node, operation_id: int, nodes_from: list):
        self.operation_id = operation_id
        if not isinstance(node.operation, str):
            # for model-based operations
            self.operation_type = node.operation.operation_type
            self.custom_params = node.parameters
            self.params = self._create_full_params(node)

            if _is_node_fitted(node):
                self.operation_name = _extract_operation_name(node)
                self._extract_fields_of_fitted_operation(node)
        else:
            # for custom operations without implementation
            self.operation_type = 'custom'
            self.custom_params = {}
            self.params = {}
            self.operation_name = node.operation
        self.nodes_from = nodes_from
        self.rating = node.rating

    def _create_full_params(self, node: Node) -> dict:
        wrapped_operations = ['base_estimator', 'estimator']

        params = {}
        if _is_node_fitted(node):
            params = extract_operation_params(node)

            # Check if it is needed to process "model in model" cases
            # such strategy is needed for RANSAC or RFE algorithms
            for wrapped_operation in wrapped_operations:
                if params is not None and wrapped_operation in params:
                    del params[wrapped_operation]

            if isinstance(self.custom_params, dict):
                for key, value in self.custom_params.items():
                    params[key] = value

        return params

    def _extract_fields_of_fitted_operation(self, node: Node):
        if 'h2o' in self.operation_type:
            self.fitted_operation_path = os.path.join('fitted_operations', f"h2o_{self.operation_id}")
        else:
            operation_name = f'operation_{str(self.operation_id)}.pkl'
            self.fitted_operation_path = os.path.join('fitted_operations', operation_name)
        self.fitted_operation = node.fitted_operation

    def convert_to_dict(self) -> dict:
        # Convert compound path into separate files and directories
        if self.fitted_operation_path is not None and os.sep in self.fitted_operation_path:
            fitted_path = os.path.split(self.fitted_operation_path)
        else:
            fitted_path = self.fitted_operation_path

        operation_object = {
            "operation_id": self.operation_id,
            "operation_type": self.operation_type,
            "operation_name": self.operation_name,
            "custom_params": self.custom_params,
            "params": self.params,
            "nodes_from": self.nodes_from,
            "fitted_operation_path": fitted_path,
            "rating": self.rating,
        }

        return operation_object

    def export_operation(self, return_path: str = None) -> Optional[Union[str, bytes, Any]]:
        """Export fitted operation: either by path or directly as bytes-like object.
        Returns None for not fitted operation.

        :param return_path: path to directory for exporting the operation.
        If None, then bytes-like object is returned directly.
        """
        if not self.fitted_operation:
            return None

        if return_path:
            check_existing_path(return_path)

            # dictionary with paths to saved fitted operations
            if 'h2o' in self.operation_type:
                self.fitted_operation_path = self.fitted_operation.save_operation(
                    os.path.join(return_path, 'fitted_operations'),
                    self.operation_id
                )
                return self.fitted_operation_path
            else:
                path_fitted_operations = os.path.join(return_path, 'fitted_operations')
                check_existing_path(path_fitted_operations)
                joblib.dump(self.fitted_operation, os.path.join(return_path, self.fitted_operation_path))
                return os.path.join(path_fitted_operations, f'operation_{self.operation_id}.pkl')
        else:
            # dictionary with bytes of fitted operations
            bytes_container = BytesIO()
            joblib.dump(self.fitted_operation, bytes_container)
            return bytes_container

    def import_json(self, operation_object: dict):
        required_fields = ['operation_id', 'operation_type', 'params', 'nodes_from']
        self._validate_json_operation_template(operation_object, required_fields)

        self.operation_id = operation_object['operation_id']
        self.operation_type = operation_object['operation_type']
        self.params = operation_object['params']
        if 'dtype' in self.params and isinstance(self.params['dtype'], str):
            self.params['dtype'] = np.dtype(self.params['dtype'])
        self.nodes_from = operation_object['nodes_from']
        if 'fitted_operation_path' in operation_object:
            fitted_operation_path = operation_object['fitted_operation_path']
            if isinstance(fitted_operation_path, str):
                self.fitted_operation_path = fitted_operation_path
            elif isinstance(fitted_operation_path, list):
                # Path were set as folders and files in tuple or list
                self.fitted_operation_path = os.path.join(*fitted_operation_path[:2])

        if 'custom_params' in operation_object:
            self.custom_params = operation_object['custom_params']
        if 'operation_name' in operation_object:
            self.operation_name = operation_object['operation_name']
        if 'rating' in operation_object:
            self.rating = operation_object['rating']


def check_existing_path(path: str):
    if not os.path.exists(path):
        os.makedirs(path)


def extract_operation_params(node: Node):
    params = node.parameters

    if 'dtype' in params:
        params['dtype'] = params['dtype'].__name__

    return params


def _extract_operation_name(node: Node):
    return node.fitted_operation.__class__.__name__


def _is_node_fitted(node: Node) -> bool:
    return bool(node.fitted_operation)



import logging
from copy import deepcopy
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from fedot.api.api_utils.api_composer import ApiComposer
from fedot.api.api_utils.api_data import ApiDataProcessor
from fedot.api.api_utils.api_data_analyser import DataAnalyser
from fedot.api.api_utils.data_definition import FeaturesType, TargetType
from fedot.api.api_utils.metrics import ApiMetrics
from fedot.api.api_utils.params import ApiParams
from fedot.api.api_utils.predefined_model import PredefinedModel
from fedot.core.constants import DEFAULT_API_TIMEOUT_MINUTES
from fedot.core.data.data import InputData, OutputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.data.visualisation import plot_biplot, plot_forecast, plot_roc_auc
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.ts_wrappers import out_of_sample_ts_forecast, convert_forecast_to_output
from fedot.core.repository.quality_metrics_repository import MetricsRepository
from fedot.core.repository.tasks import TaskParams, TaskTypesEnum
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence
from fedot.core.visualisation.opt_viz_extra import visualise_pareto
from fedot.explainability.explainer_template import Explainer
from fedot.explainability.explainers import explain_pipeline
from fedot.preprocessing.preprocessing import merge_preprocessors
from fedot.remote.remote_evaluator import RemoteEvaluator
from fedot.utilities.project_import_export import export_project_to_zip, import_project_from_zip

NOT_FITTED_ERR_MSG = 'Model not fitted yet'


class Fedot:
    """Main class for FEDOT API.

    Facade for ApiDataProcessor, ApiComposer, ApiMetrics, ApiInitialAssumptions.

    Args:
        problem: the name of modelling problem to solve

            .. details:: possible ``problem`` options:

                - ``classification`` -> for classification task
                - ``regression`` -> for regression task
                - ``ts_forecasting`` -> for time serires forecasting task

        timeout: time for model design (in minutes): ``None`` or ``-1`` means infinite time
        task_params: additional parameters of the task
        seed: value for fixed random seed
        logging_level: logging levels are the same as in 'logging'

            .. details:: possible ``logging_level`` options:

                    - ``50`` -> critical
                    - ``40`` -> error
                    - ``30`` -> warning
                    - ``20`` -> info
                    - ``10`` -> debug
                    - ``0`` -> nonset

        safe_mode: if set ``True`` it will cut large datasets to prevent memory overflow and use label encoder
            instead of oneHot encoder if summary cardinality of categorical features is high.
        n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
        max_depth: max depth of the pipeline
        max_arity: max arity of the pipeline nodes
        pop_size: population size for composer
        num_of_generations: number of generations for composer
        keep_n_best: Number of the best individuals of previous generation to keep in next generation.
        available_operations: list of model names to use
        early_stopping_iterations: composer will stop after ``n`` generation without improving
        early_stopping_timeout: stagnation timeout in minutes: composer will stop after ``n`` minutes without improving
        with_tuning: allow hyperparameters tuning for the model
        cv_folds: number of folds for cross-validation
        validation_blocks: number of validation blocks for time series forecasting
        max_pipeline_fit_time: time constraint for operation fitting (in minutes)
        initial_assumption: initial assumption for composer
        genetic_scheme: name of the genetic scheme
        history_folder: name of the folder for composing history
        metric:  metric for quality calculation during composing, also is used for tuning if ``with_tuning=True``
        collect_intermediate_metric: save metrics for intermediate (non-root) nodes in pipeline
        preset: name of preset for model building (e.g. 'best_quality', 'fast_train', 'gpu'):

            .. details:: possible ``preset`` options:

                - ``best_quality`` -> All models that are available for this data type and task are used
                - ``fast_train`` -> Models that learn quickly. This includes preprocessing operations
                  (data operations) that only reduce the dimensionality of the data, but cannot increase it.
                  For example, there are no polynomial features and one-hot encoding operations
                - ``stable`` -> The most reliable preset in which the most stable operations are included.
                - ``auto`` -> Automatically determine which preset should be used.
                - ``gpu`` -> Models that use GPU resources for computation.
                - ``ts`` -> A special preset with models for time series forecasting task.
                - ``automl`` -> A special preset with only AutoML libraries such as TPOT and H2O as operations.

        use_pipelines_cache: bool indicating whether to use pipeline structures caching, enabled by default.
        use_preprocessing_cache: bool indicating whether to use optional preprocessors caching, enabled by default.
        cache_folder: path to the place where cache files should be stored (if any cache is enabled).
        show_progress: bool indicating whether to show progress using tqdm/tuner or not
    """

    def __init__(self,
                 problem: str,
                 timeout: Optional[float] = DEFAULT_API_TIMEOUT_MINUTES,
                 task_params: TaskParams = None,
                 seed=None, logging_level: int = logging.ERROR,
                 safe_mode=False,
                 n_jobs: int = 1,
                 **composer_tuner_params
                 ):

        # Classes for dealing with metrics, data sources and hyperparameters
        self.metrics = ApiMetrics(problem)
        self.api_composer = ApiComposer(problem)
        self.params = ApiParams()

        # Define parameters, that were set via init in init
        input_params = {'problem': self.metrics.main_problem, 'timeout': timeout,
                        'composer_tuner_params': composer_tuner_params, 'task_params': task_params,
                        'seed': seed, 'logging_level': logging_level, 'n_jobs': n_jobs}
        self.params.initialize_params(input_params)

        # Initialize ApiComposer's cache parameters via ApiParams
        self.api_composer.init_cache(self.params.api_params['use_pipelines_cache'],
                                     self.params.api_params['use_preprocessing_cache'],
                                     self.params.api_params['cache_folder'])

        # Initialize data processors for data preprocessing and preliminary data analysis
        self.data_processor = ApiDataProcessor(task=self.params.api_params['task'])
        self.data_analyser = DataAnalyser(safe_mode=safe_mode)

        self.target: Optional[TargetType] = None
        self.prediction: Optional[OutputData] = None
        self._is_in_sample_prediction = True
        self.train_data: Optional[InputData] = None
        self.test_data: Optional[InputData] = None

        # Outputs
        self.current_pipeline: Optional[Pipeline] = None
        self.best_models: Sequence[Pipeline] = ()
        self.history: Optional[OptHistory] = None

    def fit(self,
            features: FeaturesType,
            target: TargetType = 'target',
            predefined_model: Union[str, Pipeline] = None) -> Pipeline:
        """Fits the graph with a predefined structure or compose and fit the new graph

        Args:
            features: the array with features of train data
            target: the array with target values of train data
            predefined_model: the name of the atomic model or Pipeline instance.
                If argument is ``auto``, perform initial assumption generation and then fit the pipeline

        Returns:
            Pipeline object

        """
        self.target = target

        self.train_data = self.data_processor.define_data(features=features, target=target, is_predict=False)

        # Launch data analyser - it gives recommendations for data preprocessing
        full_train_not_preprocessed = deepcopy(self.train_data)
        recommendations = self.data_analyser.give_recommendation(self.train_data)
        self.data_processor.accept_and_apply_recommendations(self.train_data, recommendations)
        self.params.accept_and_apply_recommendations(self.train_data, recommendations)
        self._init_remote_if_necessary()
        self.params.update_available_operations_by_preset(self.train_data)
        self.params.api_params['train_data'] = self.train_data

        if predefined_model is not None:
            # Fit predefined model and return it without composing
            self.current_pipeline = PredefinedModel(predefined_model,
                                                    self.train_data,
                                                    self.params.api_params['logger']).fit()
        else:
            self.current_pipeline, self.best_models, self.history = \
                self.api_composer.obtain_model(**self.params.api_params)

            # Final fit for obtained pipeline on full dataset
            if self.history and not self.history.is_empty() or not self.current_pipeline.is_fitted:
                self._train_pipeline_on_full_dataset(recommendations, full_train_not_preprocessed)
                self.params.api_params['logger'].message('Final pipeline was fitted')
            else:
                self.params.api_params['logger'].message('Already fitted initial pipeline is used')

        # Store data encoder in the pipeline if it is required
        self.current_pipeline.preprocessor = merge_preprocessors(self.data_processor.preprocessor,
                                                                 self.current_pipeline.preprocessor)

        self.params.api_params['logger'].message(f'Final pipeline: {self.current_pipeline.structure}')

        return self.current_pipeline

    def predict(self,
                features: FeaturesType,
                save_predictions: bool = False,
                in_sample: bool = True,
                validation_blocks: Optional[int] = None) -> np.ndarray:
        """Predicts new target using already fitted model.

        For time-series performs forecast with depth ``forecast_length`` if ``in_sample=False``.
        If ``in_sample=True`` performs in-sample forecast using features as sample.

        Args:
            features: the array with features of test data
            save_predictions: if ``True`` - save predictions as csv-file in working directory
            in_sample: used while time-series prediction. If ``in_sample=True`` performs in-sample forecast using
                features with number if iterations specified in ``validation_blocks``.
            validation_blocks: number of validation blocks for in-sample forecast.
                If ``validation_blocks = None`` uses number of validation blocks set during model initialization
                (default is 2).


        Returns:
            the array with prediction values
        """
        if self.current_pipeline is None:
            raise ValueError(NOT_FITTED_ERR_MSG)

        self.test_data = self.data_processor.define_data(target=self.target, features=features, is_predict=True)
        self._is_in_sample_prediction = in_sample
        validation_blocks = validation_blocks or self.params.api_params.get('validation_blocks')

        self.prediction = self.data_processor.define_predictions(current_pipeline=self.current_pipeline,
                                                                 test_data=self.test_data,
                                                                 in_sample=self._is_in_sample_prediction,
                                                                 validation_blocks=validation_blocks)

        if save_predictions:
            self.save_predict(self.prediction)

        return self.prediction.predict

    def predict_proba(self,
                      features: FeaturesType,
                      save_predictions: bool = False,
                      probs_for_all_classes: bool = False) -> np.ndarray:
        """Predicts the probability of new target using already fitted classification model

        Args:
            features: the array with features of test data
            save_predictions: if ``True`` - save predictions as csv-file in working directory
            probs_for_all_classes: if ``True`` - return probability for each class even for binary case

        Returns:
            the array with prediction values
        """

        if self.current_pipeline is None:
            raise ValueError(NOT_FITTED_ERR_MSG)

        if self.params.api_params['task'].task_type == TaskTypesEnum.classification:
            self.test_data = self.data_processor.define_data(target=self.target,
                                                             features=features, is_predict=True)

            mode = 'full_probs' if probs_for_all_classes else 'probs'

            self.prediction = self.current_pipeline.predict(self.test_data, output_mode=mode)

            if save_predictions:
                self.save_predict(self.prediction)
        else:
            raise ValueError('Probabilities of predictions are available only for classification')

        return self.prediction.predict

    def forecast(self,
                 pre_history: Optional[Union[str, Tuple[np.ndarray, np.ndarray], InputData, dict]] = None,
                 horizon: Optional[int] = None,
                 save_predictions: bool = False) -> np.ndarray:
        """Forecasts the new values of time series. If horizon is bigger than forecast length of fitted model -
        out-of-sample forecast is applied (not supported for multi-modal data).

        Args:
            pre_history: the array with features for pre-history of the forecast
            horizon: num of steps to forecast
            save_predictions: if ``True`` save predictions as csv-file in working directory

        Returns:
            the array with prediction values
        """
        self._check_forecast_applicable()

        forecast_length = self.train_data.task.task_params.forecast_length
        horizon = horizon if horizon is not None else forecast_length
        if pre_history is None:
            pre_history = self.train_data
            pre_history.target = None
        self.test_data = self.data_processor.define_data(target=self.target,
                                                         features=pre_history,
                                                         is_predict=True)
        predict = out_of_sample_ts_forecast(self.current_pipeline, self.test_data, horizon)
        self.prediction = convert_forecast_to_output(self.test_data, predict)
        self._is_in_sample_prediction = False
        if save_predictions:
            self.save_predict(self.prediction)
        return self.prediction.predict

    def _check_forecast_applicable(self):
        if self.current_pipeline is None:
            raise ValueError(NOT_FITTED_ERR_MSG)

        if self.params.api_params['task'].task_type != TaskTypesEnum.ts_forecasting:
            raise ValueError('Forecasting can be used only for the time series')

    def load(self, path):
        """Loads saved graph from disk

        Args:
            path: path to ``json`` file with model
        """
        self.current_pipeline = Pipeline()
        self.current_pipeline.load(path)
        self.data_processor.preprocessor = self.current_pipeline.preprocessor

    def plot_pareto(self):
        metric_names = self.params.metric_to_compose
        # archive_history stores archives of the best models.
        # Each archive is sorted from the best to the worst model,
        # so the best_candidates is sorted too.
        best_candidates = self.history.archive_history[-1]
        visualise_pareto(front=best_candidates,
                         objectives_names=metric_names,
                         show=True)

    def plot_prediction(self, in_sample: Optional[bool] = None, target: Optional[Any] = None):
        """Plots the prediction obtained from graph

        Args:
            in_sample: if current prediction is in_sample (for time-series forecasting).
            Plots predictions as future values
            target: user-specified name of target variable for :obj:`MultiModalData`

        """
        if self.prediction is not None:
            if self.params.api_params['task'].task_type == TaskTypesEnum.ts_forecasting:
                in_sample = in_sample or self._is_in_sample_prediction
                plot_forecast(self.test_data, self.prediction, in_sample, target)
            elif self.params.api_params['task'].task_type == TaskTypesEnum.regression:
                plot_biplot(self.prediction)
            elif self.params.api_params['task'].task_type == TaskTypesEnum.classification:
                self.predict_proba(self.test_data)
                plot_roc_auc(self.test_data, self.prediction)
            else:
                self.params.api_params['logger'].error('Not supported yet')
                raise NotImplementedError(f"For task {self.params.api_params['task']} plot prediction is not supported")
        else:
            self.params.api_params['logger'].error('No prediction to visualize')
            raise ValueError('Prediction from model is empty')

    def get_metrics(self,
                    target: Union[np.ndarray, pd.Series] = None,
                    metric_names: Union[str, List[str]] = None,
                    in_sample: Optional[bool] = None,
                    validation_blocks: Optional[int] = None) -> dict:
        """Gets quality metrics for the fitted graph

        Args:
            target: the array with target values of test data
            metric_names: the names of required metrics
            in_sample: used for time-series forecasting.
                If True prediction will be obtained as ``.predict(..., in_sample=True)``.
            validation_blocks: number of validation blocks for in-sample forecast.
                If ``validation_blocks = None`` uses number of validation blocks set during model initialization
                (default is 2).

        Returns:
            the values of quality metrics
        """
        if metric_names is None:
            metric_names = self.params.metric_name

        if target is not None:
            if self.test_data is None:
                self.test_data = InputData(idx=range(len(self.prediction.predict)),
                                           features=None,
                                           target=target[:len(self.prediction.predict)],
                                           task=self.train_data.task,
                                           data_type=self.train_data.data_type)
            else:
                self.test_data.target = target[:len(self.prediction.predict)]

        # TODO change to sklearn metrics
        metric_names = ensure_wrapped_in_sequence(metric_names)

        calculated_metrics = dict()
        for metric_name in metric_names:
            if self.metrics.get_metrics_mapping(metric_name) is NotImplemented:
                self.params.api_params['logger'].warning(f'{metric_name} is not available as metric')
            else:
                composer_metric = self.metrics.get_metrics_mapping(metric_name)
                metric_cls = MetricsRepository().metric_class_by_id(composer_metric)
                prediction = deepcopy(self.prediction)
                if metric_name == "roc_auc":  # for roc-auc we need probabilities
                    prediction.predict = self.predict_proba(self.test_data)
                else:
                    if in_sample is not None:
                        self._is_in_sample_prediction = in_sample
                    prediction.predict = self.predict(self.test_data, in_sample=self._is_in_sample_prediction,
                                                      validation_blocks=validation_blocks)
                real = deepcopy(self.test_data)

                # Work inplace - correct predictions
                self.data_processor.correct_predictions(real=real,
                                                        prediction=prediction)

                real.target = np.ravel(real.target)

                metric_value = abs(metric_cls.metric(reference=real,
                                                     predicted=prediction))
                calculated_metrics[metric_name] = round(metric_value, 3)

        return calculated_metrics

    def save_predict(self, predicted_data: OutputData):
        # TODO unify with OutputData.save_to_csv()
        """ Saves pipeline forecasts in csv file """
        if len(predicted_data.predict.shape) >= 2:
            prediction = predicted_data.predict.tolist()
        else:
            prediction = predicted_data.predict
        pd.DataFrame({'Index': predicted_data.idx,
                      'Prediction': prediction}).to_csv(r'./predictions.csv', index=False)
        self.params.api_params['logger'].message('Predictions was saved in current directory.')

    def export_as_project(self, project_path='fedot_project.zip'):
        export_project_to_zip(zip_name=project_path, opt_history=self.history,
                              pipeline=self.current_pipeline,
                              train_data=self.train_data, test_data=self.test_data)

    def import_as_project(self, project_path='fedot_project.zip'):
        self.current_pipeline, self.train_data, self.test_data, self.history = \
            import_project_from_zip(zip_path=project_path)
        # TODO workaround to init internal fields of API and data
        self.train_data = self.data_processor.define_data(features=self.train_data, is_predict=False)
        self.test_data = self.data_processor.define_data(features=self.test_data, is_predict=True)
        self.predict(self.test_data)

    def explain(self, features: FeaturesType = None,
                method: str = 'surrogate_dt', visualization: bool = True, **kwargs) -> 'Explainer':
        """Creates explanation for *current_pipeline* according to the selected 'method'.
            An :obj:`Explainer` instance will return.

        Args:
            features: samples to be explained. If ``None``, ``train_data`` from last fit will be used.
            method: explanation method, defaults to ``surrogate_dt``
            visualization: print and plot the explanation simultaneously, defaults to ``True``.
        Notes:
            The explanation can be retrieved later by executing :obj:`explainer.visualize()`
        """
        pipeline = self.current_pipeline
        if features is None:
            data = self.train_data
        else:
            data = self.data_processor.define_data(features=features,
                                                   is_predict=False)
        explainer = explain_pipeline(pipeline=pipeline, data=data, method=method,
                                     visualization=visualization, **kwargs)

        return explainer

    def _init_remote_if_necessary(self):
        remote = RemoteEvaluator()
        if remote.is_enabled and remote.remote_task_params is not None:
            task = self.params.api_params['task']
            if task.task_type is TaskTypesEnum.ts_forecasting:
                task_str = \
                    f'Task(TaskTypesEnum.ts_forecasting, ' \
                    f'TsForecastingParams(forecast_length={task.task_params.forecast_length}))'
            else:
                task_str = f'Task({str(task.task_type)})'
            remote.remote_task_params.task_type = task_str
            remote.remote_task_params.is_multi_modal = isinstance(self.train_data, MultiModalData)

            if isinstance(self.target, str) and remote.remote_task_params.target is None:
                remote.remote_task_params.target = self.target

    def _train_pipeline_on_full_dataset(self, recommendations: dict, full_train_not_preprocessed):
        """Applies training procedure for obtained pipeline if dataset was clipped
        """

        if recommendations:
            # if data was cut we need to refit pipeline on full data
            self.data_processor.accept_and_apply_recommendations(full_train_not_preprocessed,
                                                                 {k: v for k, v in recommendations.items()
                                                                  if k != 'cut'})
        self.current_pipeline.fit(
            full_train_not_preprocessed,
            n_jobs=self.params.api_params['n_jobs'],
        )



import itertools
from numbers import Real
from typing import Any, Optional, Iterable, Callable, Sequence, TypeVar, Dict, Tuple

from fedot.core.dag.graph import Graph
from fedot.core.log import default_log
from fedot.core.optimisers.fitness import *

G = TypeVar('G', bound=Graph, covariant=True)
R = TypeVar('R', contravariant=True)
GraphFunction = Callable[[G], R]
ObjectiveFunction = GraphFunction[G, Fitness]


class Objective:
    """Represents objective function for computing metric values
    on Graphs and keeps information about metrics used."""

    def __init__(self,
                 quality_metrics: Dict[Any, Callable],
                 complexity_metrics: Optional[Dict[Any, Callable]] = None,
                 is_multi_objective: bool = False,
                 ):
        self._log = default_log(self)
        self.is_multi_objective = is_multi_objective
        self.quality_metrics = quality_metrics
        self.complexity_metrics = complexity_metrics or {}

    def __call__(self, graph: Graph, **metrics_kwargs: Any) -> Fitness:
        evaluated_metrics = []
        for metric_id, metric_func in self.metrics:
            try:
                metric_value = metric_func(graph, **metrics_kwargs)
                evaluated_metrics.append(metric_value)
            except Exception as ex:
                self._log.error(f'Objective evaluation error for graph {graph} on metric {metric_id}: {ex}')
                return null_fitness()  # fail right away
        return to_fitness(evaluated_metrics, self.is_multi_objective)

    @property
    def metrics(self) -> Sequence[Tuple[Any, Callable]]:
        return list(itertools.chain(self.quality_metrics.items(), self.complexity_metrics.items()))

    @property
    def metric_names(self) -> Sequence[str]:
        return [str(metric_id) for metric_id, _ in self.metrics]

    def format_fitness(self, fitness: Fitness) -> str:
        """Returns formatted fitness string.
        Example for 3 metrics: `<roc_auc=0.542 f1=0.72 complexity=0.8>`"""
        fitness_info = zip(self.metric_names, fitness.values)
        fitness_info_str = [f'{name}={value:.3f}' for name, value in fitness_info]
        return f"<{' '.join(fitness_info_str)}>"


def to_fitness(metric_values: Optional[Sequence[Real]], multi_objective: bool = False) -> Fitness:
    if metric_values is None:
        return null_fitness()
    elif multi_objective:
        return MultiObjFitness(values=metric_values, weights=1.)
    else:
        return SingleObjFitness(*metric_values)



from typing import Optional

import numpy as np
import pandas as pd
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis

from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class DiscriminantAnalysisImplementation(ModelImplementation):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.model = None

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """
        self.model.fit(train_data.features, train_data.target)
        return self.model

    def predict(self, input_data):
        """ Method make prediction with labels of classes

        :param input_data: data with features to process
        """
        prediction = self.model.predict(input_data.features)

        prediction = nan_to_num(prediction)

        return prediction

    def predict_proba(self, input_data):
        """ Method make prediction with probabilities of classes

        :param input_data: data with features to process
        """
        prediction = self.model.predict_proba(input_data.features)

        prediction = nan_to_num(prediction)

        return prediction

    @property
    def classes_(self):
        return self.model.classes_


class LDAImplementation(DiscriminantAnalysisImplementation):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.model = LinearDiscriminantAnalysis(**self.params.to_dict())

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """

        self.check_and_correct_params()

        try:
            self.model.fit(train_data.features, train_data.target)
        except ValueError:
            # Problem arise when features and target are "ideally" mapping
            # features [[1.0], [0.0], [0.0]] and target [[1], [0], [0]]
            new_solver = 'lsqr'
            self.log.debug(f'Change invalid parameter solver ({self.model.solver}) to {new_solver}')

            self.model.solver = new_solver
            self.params.update(solver=new_solver)
            self.model.fit(train_data.features, train_data.target)
        return self.model

    def check_and_correct_params(self):
        """ Checks if the hyperparameters for the LDA model are correct and fixes them if needed """
        current_solver = self.params.get('solver')
        current_shrinkage = self.params.get('shrinkage')

        is_solver_svd = current_solver is not None and current_solver == 'svd'
        if is_solver_svd and current_shrinkage is not None:
            # Ignore shrinkage
            self.params.update(shrinkage=None)
            self.model.shrinkage = None


class QDAImplementation(DiscriminantAnalysisImplementation):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.model = QuadraticDiscriminantAnalysis(**self.params.to_dict())


def nan_to_num(prediction):
    """ Function converts nan values to numerical

    :return prediction: prediction without nans
    """
    if np.array([pd.isna(_) for _ in prediction]).any():
        prediction = np.nan_to_num(prediction)

    return prediction



import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc
from sklearn.model_selection import train_test_split

from examples.simple.classification.classification_pipelines import classification_random_forest_pipeline
from fedot.core.data.data import InputData
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.utilities.synth_dataset_generator import classification_dataset

np.random.seed(2020)


def get_classification_dataset(features_options, samples_amount=250,
                               features_amount=5, classes_amount=2, weights=None):
    """
    Prepares four numpy arrays with different scale features and target
    :param samples_amount: Total amount of samples in the resulted dataset.
    :param features_amount: Total amount of features per sample.
    :param classes_amount: The amount of classes in the dataset.
    :param features_options: The dictionary containing features options in key-value
    :param weights: The proportions of samples assigned to each class. If None, then classes are balanced.
    format:
        - informative: the amount of informative features;
        - redundant: the amount of redundant features;
        - repeated: the amount of features that repeat the informative features;
        - clusters_per_class: the amount of clusters for each class;
    :return x_data_train: features to train
    :return y_data_train: target to train
    :return x_data_test: features to test
    :return y_data_test: target to test
    """

    x_data, y_data = classification_dataset(samples_amount=samples_amount,
                                            features_amount=features_amount,
                                            classes_amount=classes_amount,
                                            features_options=features_options,
                                            weights=weights)

    # Changing the scale of the data
    for i, coeff in zip(range(0, features_amount),
                        np.random.randint(1, 100, features_amount)):
        # Get column
        feature = np.array(x_data[:, i])

        # Change scale for this feature
        rescaled = feature * coeff
        x_data[:, i] = rescaled

    # Train and test split
    x_data_train, x_data_test, y_data_train, y_data_test = train_test_split(x_data, y_data,
                                                                            test_size=0.3)

    return x_data_train, y_data_train, x_data_test, y_data_test


def convert_to_labels(root_operation, prediction):
    if any(root_operation == acceptable_model for acceptable_model in
           ['logit', 'lda', 'qda', 'mlp', 'svc', 'rf', 'bernb']):
        preds = np.array(prediction.predict)
        preds[preds >= 0.5] = 1
        preds[preds < 0.5] = 0
    else:
        preds = np.array(prediction.predict, dtype=int)

    return preds


def run_classification_tuning_experiment(pipeline, tuner=None):
    samples = [50, 550, 150]
    features = [1, 5, 10]
    classes = [2, 2, 2]
    options = [{'informative': 1, 'redundant': 0,
                'repeated': 0, 'clusters_per_class': 1},
               {'informative': 2, 'redundant': 1,
                'repeated': 1, 'clusters_per_class': 1},
               {'informative': 3, 'redundant': 1,
                'repeated': 2, 'clusters_per_class': 2}]

    for samples_amount, features_amount, \
        classes_amount, features_options in zip(samples, features, classes,
                                                options):
        print('=======================================')
        print(f'\nAmount of samples {samples_amount}, '
              f'amount of features {features_amount}, '
              f'amount of clsses {classes_amount}, '
              f'additional options {features_options}')

        x_train, y_train, x_test, y_test = get_classification_dataset(features_options,
                                                                      samples_amount,
                                                                      features_amount,
                                                                      classes_amount)
        task = Task(TaskTypesEnum.classification)

        # Prepare data to train the model
        train_input = InputData(idx=np.arange(0, len(x_train)),
                                features=x_train,
                                target=y_train,
                                task=task,
                                data_type=DataTypesEnum.table)

        predict_input = InputData(idx=np.arange(0, len(x_test)),
                                  features=x_test,
                                  target=None,
                                  task=task,
                                  data_type=DataTypesEnum.table)

        # Fit it
        pipeline.fit(train_input)

        # Predict
        predicted_labels = pipeline.predict(predict_input)
        preds = predicted_labels.predict

        print(f"{roc_auc(y_test, preds):.4f}\n")

        if tuner is not None:
            print('Start tuning process ...')
            pipeline_tuner = TunerBuilder(task)\
                .with_tuner(tuner)\
                .with_metric(ClassificationMetricsEnum.ROCAUC)\
                .with_iterations(50)\
                .build(train_input)
            tuned_pipeline = pipeline_tuner.tune(pipeline)

            # Fit it
            tuned_pipeline.fit(train_input)

            # Predict
            print('predict')
            predicted_values_tuned = tuned_pipeline.predict(predict_input)
            preds_tuned = predicted_values_tuned.predict

            print('Obtained metrics after tuning:')
            print(f"{roc_auc(y_test, preds_tuned):.4f}\n")


# Script for testing is pipeline can process different datasets for classification
if __name__ == '__main__':
    run_classification_tuning_experiment(pipeline=classification_random_forest_pipeline(),
                                         tuner=PipelineTuner)



import logging
import pathlib

from examples.simple.classification.classification_pipelines import classification_complex_pipeline
from examples.simple.pipeline_tune import get_case_train_test_data
from fedot.core.log import Log


def run_log_example(log_file):
    train_data, _ = get_case_train_test_data()

    # Use default_log if you do not have json config file for log
    log = Log(log_file=log_file, output_logging_level=logging.DEBUG).get_adapter(prefix=pathlib.Path(__file__).stem)

    log.info('start creating pipeline')
    pipeline = classification_complex_pipeline()

    log.info('start fitting pipeline')
    pipeline.fit(train_data)


if __name__ == '__main__':
    run_log_example(log_file='example_log.log')



from datetime import timedelta

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score as roc_auc
from sklearn.model_selection import train_test_split

from examples.simple.classification.classification_pipelines import classification_pipeline_without_balancing, \
    classification_pipeline_with_balancing
from examples.simple.classification.classification_with_tuning import get_classification_dataset
from fedot.core.data.data import InputData
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import TaskTypesEnum, Task
from fedot.core.utils import fedot_project_root


def run_resample_example(path_to_data=None, tune=False):
    if path_to_data is None:
        samples = 1000
        features = 10
        classes = 2
        weights = [0.45, 0.55]
        features_options = {'informative': 1, 'redundant': 1, 'repeated': 1, 'clusters_per_class': 1}

        x_train, y_train, x_test, y_test = get_classification_dataset(features_options,
                                                                      samples,
                                                                      features,
                                                                      classes,
                                                                      weights)
    else:
        data = pd.read_csv(path_to_data, header=0)

        features = data.drop(columns='Class')
        target = data['Class']

        x_train, x_test, y_train, y_test = train_test_split(np.array(features),
                                                            np.array(target), test_size=0.3)

    unique_class, counts_class = np.unique(y_train, return_counts=True)
    print(f'Two classes: {unique_class}')
    print(f'{unique_class[0]}: {counts_class[0]}')
    print(f'{unique_class[1]}: {counts_class[1]}')

    task = Task(TaskTypesEnum.classification)

    train_input = InputData(idx=np.arange(0, len(x_train)),
                            features=x_train,
                            target=y_train,
                            task=task,
                            data_type=DataTypesEnum.table)

    predict_input = InputData(idx=np.arange(0, len(x_test)),
                              features=x_test,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.table)

    print('Begin fit Pipeline without balancing')
    # Pipeline without balancing
    pipeline = classification_pipeline_without_balancing()
    pipeline.fit_from_scratch(train_input)

    # Predict
    predict_labels = pipeline.predict(predict_input)
    preds = predict_labels.predict
    print('---')
    print(f'ROC-AUC of pipeline without balancing {roc_auc(y_test, preds):.4f}\n')

    # Pipeline with balancing
    pipeline = classification_pipeline_with_balancing()

    print('Begin fit Pipeline with balancing')
    # pipeline.fit(train_input)
    pipeline.fit(train_input)

    # Predict
    predict_labels = pipeline.predict(predict_input)
    preds = predict_labels.predict
    print('---')
    print(f'ROC-AUC of pipeline with balancing {roc_auc(y_test, preds):.4f}\n')

    if tune:
        print('Start tuning process ...')
        tuner = TunerBuilder(train_input.task)\
            .with_tuner(PipelineTuner)\
            .with_metric(RegressionMetricsEnum.MAE)\
            .with_iterations(50) \
            .with_timeout(timedelta(minutes=1))\
            .build(train_input)
        tuned_pipeline = tuner.tune(pipeline)
        # Fit
        pipeline.fit(train_input)
        # Predict
        predicted_values_tuned = tuned_pipeline.predict(predict_input)
        preds_tuned = predicted_values_tuned.predict

        print('Obtained metrics after tuning:')
        print(f'ROC-AUC of tuned pipeline with balancing - {roc_auc(y_test, preds_tuned):.4f}\n')


if __name__ == '__main__':
    run_resample_example()
    print('=' * 25)
    run_resample_example(f'{fedot_project_root()}/examples/data/credit_card_anomaly.csv', tune=True)



import os
from datetime import timedelta
from typing import Optional, Callable
from uuid import uuid4

from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.remote.infrastructure.clients.client import Client
from fedot.remote.run_pipeline import fit_pipeline


class TestClient(Client):
    def __init__(self, connect_params: dict, exec_params: dict, output_path: Optional[str] = None):
        self.connect_params = connect_params
        self.exec_params = exec_params
        self.output_path = output_path if output_path else \
            os.path.join(default_fedot_data_dir(), 'remote_fit_results')
        super().__init__(connect_params, exec_params, output_path)

    def create_task(self, config) -> str:
        fit_pipeline(config)
        return str(uuid4())

    def wait_until_ready(self) -> timedelta:
        return timedelta()

    def download_result(self, execution_id: int, result_cls=Pipeline) -> Pipeline:
        results_path_out = os.path.join(self.output_path)
        results_folder = os.listdir(results_path_out)[0]
        pipeline = result_cls.from_serialized(os.path.join(results_path_out, results_folder, 'fitted_pipeline.json'))
        return pipeline



import traceback
from typing import Union

from fedot.api.api_utils.assumptions.assumptions_builder import AssumptionsBuilder
from fedot.core.data.data import InputData
from fedot.core.log import Log
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.verification import verify_pipeline, verifier_for_task


class PredefinedModel:
    def __init__(self, predefined_model: Union[str, Pipeline], data: InputData, log: Log):
        self.predefined_model = predefined_model
        self.data = data
        self.log = log
        self.pipeline = self._get_pipeline()

    def _get_pipeline(self) -> Pipeline:
        if isinstance(self.predefined_model, Pipeline):
            pipelines = self.predefined_model
        elif self.predefined_model == 'auto':
            # Generate initial assumption automatically
            pipelines = AssumptionsBuilder.get(self.data).from_operations().build()[0]
        elif isinstance(self.predefined_model, str):
            model = PrimaryNode(self.predefined_model)
            pipelines = Pipeline(model)
        else:
            raise ValueError(f'{type(self.predefined_model)} is not supported as Fedot model')

        verify_pipeline(pipelines, task_type=self.data.task.task_type, raise_on_failure=True)

        return pipelines

    def fit(self):
        try:
            self.pipeline.fit(self.data)
        except Exception as ex:
            fit_failed_info = f'Predefined model fit was failed due to: {ex}.'
            advice_info = f'{fit_failed_info} Check pipeline structure and the correctness of the data'
            self.log.message(fit_failed_info)
            print(traceback.format_exc())
            raise ValueError(advice_info)
        return self.pipeline



import datetime
from dataclasses import dataclass, field
from typing import Optional

from fedot.core.utils import default_fedot_data_dir


@dataclass
class ComposerRequirements:
    """Defines general algorithm-independent parameters of the composition process
    (like stop condition, validation, timeout, logging etc.)

    Options related to stop condition:
    :param num_of_generations: maximum number of optimizer generations
    :param timeout: max time in minutes available for composition process
    :param early_stopping_iterations: optional max number of stagnating
    populations for early stopping. If None -- do not use early stopping.

    Infrastructure options (logging, performance)
    :param keep_n_best: number of the best individuals of previous generation to keep in next generation
    :param max_pipeline_fit_time: time constraint for operation fitting (minutes)
    :param n_jobs: num of n_jobs
    :param show_progress: bool indicating whether to show progress using tqdm or not
    :param collect_intermediate_metric: save metrics for intermediate (non-root) nodes in pipeline

    History options:
    :param keep_history: if True, then save generations to history; if False, don't keep history.
    :param history_dir: directory for saving optimization history, optional.
      If the path is relative, then save relative to `default_fedot_data_dir`.
      If absolute -- then save directly by specified path.
      If None -- do not save the history to disk and keep it only in-memory.

    Model validation options:
    :param cv_folds: number of cross-validation folds
    :param validation_blocks: number of validation blocks for time series validation
    """

    num_of_generations: Optional[int] = None
    timeout: Optional[datetime.timedelta] = datetime.timedelta(minutes=5)
    early_stopping_iterations: Optional[int] = 10
    early_stopping_timeout: Optional[float] = 5
    keep_n_best: int = 1
    max_pipeline_fit_time: Optional[datetime.timedelta] = None
    n_jobs: int = 1
    show_progress: bool = True
    collect_intermediate_metric: bool = False

    keep_history: bool = True
    history_dir: Optional[str] = field(default_factory=default_fedot_data_dir)

    cv_folds: Optional[int] = None
    validation_blocks: Optional[int] = None

    def __post_init__(self):
        if self.cv_folds is not None and self.cv_folds <= 1:
            raise ValueError('Number of folds for KFold cross validation must be 2 or more.')



import datetime
from typing import Any, List

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error

from examples.simple.time_series_forecasting.ts_pipelines import ts_complex_ridge_pipeline
from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.composer.gp_composer.specific_operators import parameter_change_mutation
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.mutation import MutationTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import \
    RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import fedot_project_root


def get_available_operations():
    """ Function returns available operations for primary and secondary nodes """
    primary_operations = ['lagged', 'smoothing', 'gaussian_filter', 'ar']
    secondary_operations = ['lagged', 'ridge', 'lasso', 'knnreg', 'linear',
                            'scaling', 'ransac_lin_reg', 'rfe_lin_reg']
    return primary_operations, secondary_operations


datasets = {
    'australia': f'{fedot_project_root()}/examples/data/ts/australia.csv',
    'beer': f'{fedot_project_root()}/examples/data/ts/beer.csv',
    'salaries': f'{fedot_project_root()}/examples/data/ts/salaries.csv',
    'stackoverflow': f'{fedot_project_root()}/examples/data/ts/stackoverflow.csv'}


def run_composing(dataset: str, pipeline: Pipeline, len_forecast=250):
    """ Example of ts forecasting using custom pipelines with composing
    :param dataset: name of dataset
    :param pipeline: pipeline to use
    :param len_forecast: forecast length
    """
    # show initial pipeline
    pipeline.print_structure()

    time_series = pd.read_csv(datasets[dataset])
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=len_forecast))
    if dataset not in ['australia']:
        idx = pd.to_datetime(time_series['idx'].values)
    else:
        # non datetime indexes
        idx = time_series['idx'].values
    time_series = time_series['value'].values
    train_input = InputData(idx=idx,
                            features=time_series,
                            target=time_series,
                            task=task,
                            data_type=DataTypesEnum.ts)
    train_data, test_data = train_test_data_setup(train_input)
    test_target = np.ravel(test_data.target)

    pipeline.fit(train_data)

    prediction = pipeline.predict(test_data)
    predict = np.ravel(np.array(prediction.predict))

    plot_info = []
    metrics_info = {}
    plot_info.append({'idx': idx,
                      'series': time_series,
                      'label': 'Actual time series'})

    rmse = mean_squared_error(test_target, predict, squared=False)
    mae = mean_absolute_error(test_target, predict)

    metrics_info['Metrics without composing'] = {'RMSE': round(rmse, 3),
                                                 'MAE': round(mae, 3)}
    plot_info.append({'idx': prediction.idx,
                      'series': predict,
                      'label': 'Forecast without composing'})
    plot_info.append(get_border_line_info(prediction.idx[0], predict, time_series, 'Border line'))

    # Get available_operations type
    primary_operations, secondary_operations = get_available_operations()

    # Composer parameters
    composer_requirements = PipelineComposerRequirements(
        primary=primary_operations,
        secondary=secondary_operations,
        max_arity=3, max_depth=8,
        num_of_generations=10,
        timeout=datetime.timedelta(minutes=10),
        cv_folds=2,
        validation_blocks=2
    )
    optimizer_parameters = GPGraphOptimizerParameters(
        pop_size=10,
        crossover_prob=0.8, mutation_prob=0.8,
        mutation_types=[parameter_change_mutation,
                        MutationTypesEnum.growth,
                        MutationTypesEnum.reduce,
                        MutationTypesEnum.simple]
    )
    composer = ComposerBuilder(task). \
        with_requirements(composer_requirements). \
        with_optimizer_params(optimizer_parameters). \
        with_metrics(RegressionMetricsEnum.RMSE). \
        with_initial_pipelines([pipeline]). \
        build()

    obtained_pipeline = composer.compose_pipeline(data=train_data)

    obtained_pipeline.fit_from_scratch(train_data)
    prediction_after = obtained_pipeline.predict(test_data)
    predict_after = np.ravel(np.array(prediction_after.predict))

    rmse = mean_squared_error(test_target, predict_after, squared=False)
    mae = mean_absolute_error(test_target, predict_after)

    metrics_info['Metrics after composing'] = {'RMSE': round(rmse, 3),
                                               'MAE': round(mae, 3)}
    plot_info.append({'idx': prediction_after.idx,
                      'series': predict_after,
                      'label': 'Forecast after composing'})
    print(metrics_info)

    visualise(plot_info)
    # structure of obtained pipeline
    obtained_pipeline.print_structure()
    obtained_pipeline.show()


def visualise(plot_info: List[dict]):
    """
    Creates a plot based on plot_info

    :param plot_info: list of parameters for plot:
    The possible parameters are:
            'idx' - idx (or x axis data)
            'series' - data to plot (or y axis data)
            'label' - label for legend
            'color' - (optional) color of line
    """
    plt.figure()
    for p in plot_info:
        color = p.get('color')
        plt.plot(p['idx'], p['series'], label=p['label'], color=color)
    plt.legend()
    plt.grid()
    plt.show()


def get_border_line_info(idx: Any, predict: np.array, time_series: np.array, label: str, color: str = 'black') -> dict:
    """
    Return plot_info for border vertical line that divides train and test part of data

    :param idx: idx for vertical line
    :param predict: predictions
    :param time_series: full time series with test_data
    :param label: label for a legend
    :parma color: color of a line
    """
    return {'idx': [idx, idx],
            'series': [min(np.concatenate([np.ravel(time_series), predict])),
                       max(np.concatenate([np.ravel(time_series), predict]))],
            'label': label,
            'color': color}


if __name__ == '__main__':
    run_composing('australia', ts_complex_ridge_pipeline(), len_forecast=10)



from copy import deepcopy
from os.path import join
from threading import Lock, Thread
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import mean_squared_error

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.operations.operation_template import extract_operation_params
from fedot.core.pipelines.node import Node
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.node_sa_approaches import NodeAnalyzeApproach
from fedot.sensitivity.operations_hp_sensitivity.problem import OneOperationProblem
from fedot.sensitivity.operations_hp_sensitivity.sa_and_sample_methods import (
    analyze_method_by_name,
    sample_method_by_name
)
from fedot.sensitivity.sa_requirements import HyperparamsAnalysisMetaParams, SensitivityAnalysisRequirements


class OneOperationHPAnalyze(NodeAnalyzeApproach):
    lock = Lock()

    def __init__(self, pipeline: Pipeline, train_data, test_data: InputData,
                 requirements: SensitivityAnalysisRequirements = None, path_to_save=None):
        super().__init__(pipeline, train_data, test_data, path_to_save)

        requirements = SensitivityAnalysisRequirements() if requirements is None else requirements
        self.requirements: HyperparamsAnalysisMetaParams = requirements.hp_analysis_meta

        self.analyze_method = analyze_method_by_name.get(self.requirements.analyze_method)
        self.sample_method = sample_method_by_name.get(self.requirements.sample_method)
        self.problem = None
        self.operation_type = None
        self.data_under_lock: dict = {}

        self.path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'nodes_sensitivity') if path_to_save is None else path_to_save
        self.log = default_log(self)

    def analyze(self, node: Node,
                is_dispersion_analysis: bool = False) -> Union[dict, float]:

        # check whether the pipeline is fitted
        if not self._pipeline.is_fitted:
            self._pipeline.fit(self._train_data)

        # create problem
        self.operation_type = node.operation.operation_type
        self.problem = OneOperationProblem(operation_types=[self.operation_type])

        # sample
        samples = self.sample(self.requirements.sample_size, node)

        response_matrix = self._get_response_matrix(samples)
        indices = self.analyze_method(self.problem.dictionary, samples, response_matrix)
        converted_to_json_indices = self._convert_indices_to_json(problem=self.problem,
                                                                  si=indices)

        # Analyze the dispersion of params
        if is_dispersion_analysis:
            self._dispersion_analysis(node=node,
                                      sample_size=self.requirements.sample_size)

        return converted_to_json_indices

    def sample(self, *args) -> Union[List[Pipeline], Pipeline]:
        """
        """

        sample_size, node = args
        samples: List[np.array] = self.sample_method(self.problem.dictionary, num_of_samples=sample_size)

        converted_samples: List[dict] = self.problem.convert_sample_to_dict(samples)
        sampled_pipelines: List[Pipeline] = self._apply_params_to_node(params=converted_samples,
                                                                       node=node)

        return sampled_pipelines

    def _apply_params_to_node(self, params: List[dict], node: Node) -> List[Pipeline]:
        sampled_pipelines: List[Pipeline] = list()
        for sample in params:
            copied_pipeline = deepcopy(self._pipeline)
            node_id = self._pipeline.nodes.index(node)
            copied_pipeline.nodes[node_id].parameters = sample
            sampled_pipelines.append(copied_pipeline)

        return sampled_pipelines

    def _get_response_matrix(self, samples: List[Pipeline]):
        operation_response_matrix = []
        for sampled_pipeline in samples:
            sampled_pipeline.fit(self._train_data)
            prediction = sampled_pipeline.predict(self._test_data)
            mse_metric = mean_squared_error(y_true=self._test_data.target,
                                            y_pred=prediction.predict)
            operation_response_matrix.append(mse_metric)

        return np.array(operation_response_matrix)

    def _dispersion_analysis(self, node: Node, sample_size: int):
        samples: np.array = self.sample_method(self.problem.dictionary, num_of_samples=sample_size)
        transposed_samples = samples.T
        converted_samples = self.problem.convert_for_dispersion_analysis(transposed_samples)

        jobs = [Thread(target=self._evaluate_variance,
                       args=(params, transposed_samples[index], node))
                for index, params in enumerate(converted_samples)]

        for job in jobs:
            job.start()

        for job in jobs:
            job.join()

        self._visualize_variance()

    def _evaluate_variance(self, params: List[dict], samples, node: Node):
        # default values of param & loss
        param_name = list(params[0].keys())[0]
        default_param_value = extract_operation_params(node).get(param_name)
        pipelines_with_applied_params = self._apply_params_to_node(params, node)

        # percentage ratio
        samples = (samples - default_param_value) / default_param_value
        response_matrix = self._get_response_matrix(pipelines_with_applied_params)
        response_matrix = (response_matrix - np.mean(response_matrix)) / \
                          (max(response_matrix) - min(response_matrix))

        OneOperationHPAnalyze.lock.acquire()
        self.data_under_lock[f'{param_name}'] = [samples.reshape(1, -1)[0], response_matrix]
        OneOperationHPAnalyze.lock.release()

    def _visualize_variance(self):
        x_ticks_param = list()
        x_ticks_loss = list()
        for param in self.data_under_lock.keys():
            x_ticks_param.append(param)
            x_ticks_loss.append(f'{param}_loss')
        param_values_data = list()
        losses_data = list()
        for value in self.data_under_lock.values():
            param_values_data.append(value[0])
            losses_data.append(value[1])

        fig, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))
        ax1.boxplot(param_values_data)
        ax2.boxplot(losses_data)
        ax1.set_title('param')
        ax1.set_xticks(range(1, len(x_ticks_param) + 1))
        ax1.set_xticklabels(x_ticks_param)
        ax2.set_title('loss')
        ax2.set_xticks(range(1, len(x_ticks_loss) + 1))
        ax2.set_xticklabels(x_ticks_loss)

        plt.savefig(join(self._path_to_save, f'{self.operation_type}_hp_sa.jpg'))

    @staticmethod
    def _convert_indices_to_json(problem: OneOperationProblem, si: dict) -> dict:
        sobol_indices = []
        for index in range(problem.num_vars):
            var_indices = {f"{problem.names[index]}": {
                'S1': list(si['S1'])[index],
                'S1_conf': list(si['S1_conf'])[index],
                'ST': list(si['ST'])[index],
                'ST_conf': list(si['ST_conf'])[index],
            }}
            sobol_indices.append(var_indices)

        data = {
            'problem': problem.dictionary,
            'sobol_indices': sobol_indices
        }

        return data



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from hyperopt import hp
from sklearn.linear_model import Ridge

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.search_space import SearchSpace
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import TaskTypesEnum, Task, TsForecastingParams


# implementation of custom model without fitting
def domain_model_imitation_predict(fitted_model: any, idx: np.array, predict_data: np.array, params: dict):
    # TODO real custom model or more realistic imitation
    a = params.get('a')
    b = params.get('b')
    shape = predict_data.shape
    result = np.random.rand(*shape) * a + b
    # Available output_type's 'table', 'ts', 'image', 'text'
    return result, 'table'


# implementation of custom regression model imitation (fit)
def custom_ml_model_imitation_fit(idx: np.array, features: np.array, target: np.array, params: dict):
    alpha = params.get('alpha')
    reg = Ridge(alpha=alpha)
    reg.fit(features, target)
    return reg


# implementation of custom regression model imitation (predict)
def custom_ml_model_imitation_predict(fitted_model: any, idx: np.array, features: np.array, params: dict):
    res = fitted_model.predict(features)
    return res, 'table'


def get_domain_pipeline():
    """
        Pipeline looking like this
        lagged -> custom -> ridge
    """
    lagged_node = PrimaryNode('lagged')
    lagged_node.parameters = {'window_size': 50}

    # For custom model params as initial approximation and model as function is necessary
    custom_node = SecondaryNode('custom', nodes_from=[lagged_node])
    custom_node.parameters = {"a": -50, "b": 500, 'model_predict': domain_model_imitation_predict}

    node_final = SecondaryNode('ridge', nodes_from=[custom_node])
    pipeline = Pipeline(node_final)

    return pipeline


def get_fitting_custom_pipeline():
    """
        Pipeline looking like this
        lagged -> custom -> ridge
    """
    lagged_node = PrimaryNode('lagged')
    lagged_node.parameters = {'window_size': 50}

    # For custom model params as initial approximation and model as function is necessary
    custom_node = SecondaryNode('custom', nodes_from=[lagged_node])
    custom_node.parameters = {'alpha': 5,
                                 'model_predict': custom_ml_model_imitation_predict,
                                 'model_fit': custom_ml_model_imitation_fit}

    node_final = SecondaryNode('lasso', nodes_from=[custom_node])
    pipeline = Pipeline(node_final)

    return pipeline


def run_pipeline_tuning(time_series, len_forecast, pipeline_type):
    # Source time series
    train_input, predict_input = train_test_data_setup(InputData(idx=range(len(time_series)),
                                                                 features=time_series,
                                                                 target=time_series,
                                                                 task=Task(TaskTypesEnum.ts_forecasting,
                                                                           TsForecastingParams(
                                                                               forecast_length=len_forecast)),
                                                                 data_type=DataTypesEnum.ts))

    if pipeline_type == 'with_fit':
        pipeline = get_fitting_custom_pipeline()
        # Setting custom search space for tuner (necessary)
        # model and output_type should be wrapped into hyperopt
        custom_search_space = {'custom': {
            'alpha': (hp.uniform, [0.01, 10]),
            'model_predict': (hp.choice, [[custom_ml_model_imitation_predict]]),
            'model_fit': (hp.choice, [[custom_ml_model_imitation_fit]])}}
    elif pipeline_type == 'without_fit':
        pipeline = get_domain_pipeline()
        # Setting custom search space for tuner (necessary)
        # model and output_type should be wrapped into hyperopt
        custom_search_space = {'custom': {'a': (hp.uniform, [-100, 100]),
                                          'b': (hp.uniform, [0, 1000]),
                                          'model_predict': (hp.choice, [[domain_model_imitation_predict]])}}
    pipeline.fit_from_scratch(train_input)
    pipeline.print_structure()
    # Get prediction with initial approximation
    predicted_before_tuning = pipeline.predict(predict_input).predict

    replace_default_search_space = True
    cv_folds = 3
    validation_blocks = 3
    search_space = SearchSpace(custom_search_space=custom_search_space,
                               replace_default_search_space=replace_default_search_space)
    pipeline_tuner = TunerBuilder(train_input.task)\
        .with_tuner(PipelineTuner)\
        .with_metric(RegressionMetricsEnum.RMSE)\
        .with_cv_folds(cv_folds)\
        .with_validation_blocks(validation_blocks)\
        .with_iterations(10)\
        .with_search_space(search_space).build(train_input)
    # Tuning pipeline
    pipeline = pipeline_tuner.tune(pipeline)
    # Fit pipeline on the entire train data
    pipeline.fit_from_scratch(train_input)
    # Predict tuned pipeline
    predicted_values = pipeline.predict(predict_input).predict
    pipeline.print_structure()

    plt.plot(np.arange(len(predicted_before_tuning[0])), predicted_before_tuning[0], label='Before tuning')
    plt.plot(np.arange(len(predicted_values[0])), predicted_values[0], label='After tuning')
    plt.plot(np.arange(len(predict_input.target)), predict_input.target, label='Real')
    plt.legend()
    plt.show()


if __name__ == '__main__':
    df = pd.read_csv('../../../cases/data/time_series/metocean.csv')
    time_series = np.array(df['value'])
    run_pipeline_tuning(time_series=time_series,
                        len_forecast=50,
                        pipeline_type='with_fit')  # mean custom ml model with fit, 'without_fit' - means domain model



import datetime
import gc
import os
from typing import Callable, List, Optional, Sequence, Tuple, Union

from fedot.api.api_utils.assumptions.assumptions_handler import AssumptionsHandler
from fedot.api.api_utils.metrics import ApiMetrics
from fedot.api.api_utils.presets import OperationsPreset
from fedot.api.time import ApiTime
from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.pipelines.pipeline_advisor import PipelineChangeAdvisor
from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.composer.gp_composer.gp_composer import GPComposer
from fedot.core.composer.gp_composer.specific_operators import boosting_mutation, parameter_change_mutation
from fedot.core.constants import DEFAULT_TUNING_ITERATIONS_NUMBER
from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.adapters import PipelineAdapter
from fedot.core.optimisers.gp_comp.evaluation import determine_n_jobs
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.inheritance import GeneticSchemeTypesEnum
from fedot.core.optimisers.gp_comp.operators.mutation import MutationTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.optimisers.optimizer import GraphGenerationParams
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_node_factory import PipelineOptNodeFactory
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.pipelines.verification import rules_by_task
from fedot.core.repository.pipeline_operation_repository import PipelineOperationRepository
from fedot.core.repository.quality_metrics_repository import MetricsRepository, MetricType, MetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence
from fedot.utilities.define_metric_by_task import MetricByTask


class ApiComposer:

    def __init__(self, problem: str):
        self.log = default_log(self)
        self.metrics = ApiMetrics(problem)
        self.pipelines_cache: Optional[OperationsCache] = None
        self.preprocessing_cache: Optional[PreprocessingCache] = None
        self.preset_name = None
        self.timer = None
        # status flag indicating that composer step was applied
        self.was_optimised = False
        # status flag indicating that tuner step was applied
        self.was_tuned = False

    def obtain_metric(self, task: Task, metric: Union[str, MetricsEnum, Callable, Sequence]) -> Sequence[MetricType]:
        """Chooses metric to use for quality assessment of pipeline during composition"""
        if metric is None:
            metric = MetricByTask.get_default_quality_metrics(task.task_type)

        metric_ids = []
        for specific_metric in ensure_wrapped_in_sequence(metric):
            if isinstance(specific_metric, Callable):
                metric = specific_metric
            else:
                metric = None
                if isinstance(specific_metric, str):
                    # Composer metric was defined by name (str)
                    metric = self.metrics.get_metrics_mapping(metric_name=specific_metric)
                elif isinstance(specific_metric, MetricsEnum):
                    metric = specific_metric
            if metric is None:
                raise ValueError(f'Incorrect metric {specific_metric}')
            metric_ids.append(metric)
        return metric_ids

    def obtain_model(self, **common_dict):
        # Prepare parameters
        api_params_dict, composer_params_dict, tuner_params_dict = _divide_parameters(common_dict)
        # Start composing - pipeline structure search
        return self.compose_fedot_model(api_params_dict, composer_params_dict, tuner_params_dict)

    def init_cache(self, use_pipelines_cache: bool, use_preprocessing_cache: bool,
                   cache_folder: Optional[Union[str, os.PathLike]] = None):
        if use_pipelines_cache:
            self.pipelines_cache = OperationsCache(cache_folder)
            #  in case of previously generated singleton cache
            self.pipelines_cache.reset()
        if use_preprocessing_cache:
            self.preprocessing_cache = PreprocessingCache(cache_folder)
            #  in case of previously generated singleton cache
            self.preprocessing_cache.reset()

    @staticmethod
    def _init_composer_requirements(api_params: dict,
                                    composer_params: dict,
                                    datetime_composing: Optional[datetime.timedelta],
                                    preset: str) -> PipelineComposerRequirements:

        task = api_params['task']

        # define available operations
        available_operations = composer_params.get('available_operations',
                                                   OperationsPreset(task, preset).filter_operations_by_preset())
        primary_operations, secondary_operations = \
            PipelineOperationRepository.divide_operations(available_operations, task)

        max_pipeline_fit_time = composer_params['max_pipeline_fit_time']
        if max_pipeline_fit_time:
            max_pipeline_fit_time = datetime.timedelta(minutes=max_pipeline_fit_time)

        composer_requirements = PipelineComposerRequirements(
            primary=primary_operations,
            secondary=secondary_operations,
            max_arity=composer_params['max_arity'],
            max_depth=composer_params['max_depth'],

            num_of_generations=composer_params['num_of_generations'],
            timeout=datetime_composing,
            early_stopping_iterations=composer_params.get('early_stopping_iterations', None),
            early_stopping_timeout=composer_params.get('early_stopping_timeout', None),
            max_pipeline_fit_time=max_pipeline_fit_time,
            n_jobs=api_params['n_jobs'],
            show_progress=api_params['show_progress'],
            collect_intermediate_metric=composer_params['collect_intermediate_metric'],
            keep_n_best=composer_params['keep_n_best'],

            keep_history=True,
            history_dir=composer_params.get('history_folder'),

            cv_folds=composer_params['cv_folds'],
            validation_blocks=composer_params['validation_blocks'],
        )
        return composer_requirements

    @staticmethod
    def _init_optimizer_parameters(composer_params: dict,
                                   multi_objective: bool,
                                   task_type: TaskTypesEnum) -> GPGraphOptimizerParameters:
        genetic_scheme_type = GeneticSchemeTypesEnum.parameter_free
        if composer_params['genetic_scheme'] == 'steady_state':
            genetic_scheme_type = GeneticSchemeTypesEnum.steady_state

        optimizer_params = GPGraphOptimizerParameters(
            multi_objective=multi_objective,
            pop_size=composer_params['pop_size'],
            genetic_scheme_type=genetic_scheme_type,
            mutation_types=ApiComposer._get_default_mutations(task_type)
        )
        return optimizer_params

    @staticmethod
    def _get_default_mutations(task_type: TaskTypesEnum) -> Sequence[MutationTypesEnum]:
        mutations = [parameter_change_mutation,
                     MutationTypesEnum.single_change,
                     MutationTypesEnum.single_drop,
                     MutationTypesEnum.single_add]

        # TODO remove workaround after boosting mutation fix
        if task_type == TaskTypesEnum.ts_forecasting:
            mutations.append(boosting_mutation)
        # TODO remove workaround after validation fix
        if task_type is not TaskTypesEnum.ts_forecasting:
            mutations.append(MutationTypesEnum.single_edge)

        return mutations

    @staticmethod
    def _init_graph_generation_params(task: Task, preset: str, available_operations: List[str],
                                      requirements: PipelineComposerRequirements):
        advisor = PipelineChangeAdvisor(task)
        graph_model_repo = PipelineOperationRepository()\
            .from_available_operations(task=task, preset=preset,
                                       available_operations=available_operations)
        node_factory = PipelineOptNodeFactory(requirements=requirements, advisor=advisor,
                                              graph_model_repository=graph_model_repo) \
            if requirements else None
        graph_generation_params = GraphGenerationParams(adapter=PipelineAdapter(),
                                                        rules_for_constraint=rules_by_task(task.task_type),
                                                        advisor=advisor,
                                                        node_factory=node_factory)
        return graph_generation_params

    def compose_fedot_model(self, api_params: dict, composer_params: dict, tuning_params: dict) \
            -> Tuple[Pipeline, Sequence[Pipeline], OptHistory]:
        """ Function for composing FEDOT pipeline model """
        task: Task = api_params['task']
        train_data = api_params['train_data']
        timeout = api_params['timeout']
        with_tuning = tuning_params['with_tuning']
        available_operations = composer_params['available_operations']
        preset = composer_params['preset']

        self.timer = ApiTime(time_for_automl=timeout, with_tuning=with_tuning)

        # Work with initial assumptions
        assumption_handler = AssumptionsHandler(train_data)

        initial_assumption = assumption_handler.propose_assumptions(composer_params['initial_assumption'],
                                                                    available_operations)
        with self.timer.launch_assumption_fit():
            fitted_assumption = \
                assumption_handler.fit_assumption_and_check_correctness(initial_assumption[0],
                                                                        pipelines_cache=self.pipelines_cache,
                                                                        preprocessing_cache=self.preprocessing_cache)

        self.log.message(
            f'Initial pipeline was fitted in {round(self.timer.assumption_fit_spend_time.total_seconds(), 1)} sec.')

        n_jobs = determine_n_jobs(api_params['n_jobs'])
        self.preset_name = assumption_handler.propose_preset(preset, self.timer, n_jobs=n_jobs)

        composer_requirements = ApiComposer._init_composer_requirements(api_params, composer_params,
                                                                        self.timer.timedelta_composing,
                                                                        self.preset_name)

        metric_functions = self.obtain_metric(task, composer_params['metric'])
        graph_generation_params = \
            self._init_graph_generation_params(task=task,
                                               preset=preset,
                                               available_operations=composer_params.get('available_operations'),
                                               requirements=composer_requirements)
        self.log.message(f"AutoML configured."
                         f" Parameters tuning: {with_tuning}."
                         f" Time limit: {timeout} min."
                         f" Set of candidate models: {available_operations}.")

        best_pipeline, best_pipeline_candidates, gp_composer = self.compose_pipeline(task, train_data,
                                                                                     fitted_assumption,
                                                                                     metric_functions,
                                                                                     composer_requirements,
                                                                                     composer_params,
                                                                                     graph_generation_params)
        if with_tuning:
            best_pipeline = self.tune_final_pipeline(task, train_data,
                                                     metric_functions[0],
                                                     composer_requirements,
                                                     best_pipeline)
        if gp_composer.history:
            adapter = gp_composer.optimizer.graph_generation_params.adapter
            gp_composer.history.tuning_result = adapter.adapt(best_pipeline)
        # enforce memory cleaning
        gc.collect()

        self.log.message('Model generation finished')
        return best_pipeline, best_pipeline_candidates, gp_composer.history

    def compose_pipeline(self, task: Task,
                         train_data: InputData,
                         fitted_assumption: Pipeline,
                         metric_functions: Sequence[MetricsEnum],
                         composer_requirements: PipelineComposerRequirements,
                         composer_params: dict,
                         graph_generation_params: GraphGenerationParams,
                         ) -> Tuple[Pipeline, List[Pipeline], GPComposer]:

        multi_objective = len(metric_functions) > 1
        optimizer_params = ApiComposer._init_optimizer_parameters(composer_params,
                                                                  multi_objective=multi_objective,
                                                                  task_type=task.task_type)

        gp_composer: GPComposer = ComposerBuilder(task=task) \
            .with_requirements(composer_requirements) \
            .with_initial_pipelines(fitted_assumption) \
            .with_optimizer(composer_params.get('optimizer')) \
            .with_optimizer_params(parameters=optimizer_params,
                                   external_parameters=composer_params.get('optimizer_external_params')) \
            .with_metrics(metric_functions) \
            .with_cache(self.pipelines_cache, self.preprocessing_cache) \
            .with_graph_generation_param(graph_generation_params=graph_generation_params) \
            .build()

        n_jobs = determine_n_jobs(composer_requirements.n_jobs)
        if self.timer.have_time_for_composing(composer_params['pop_size'], n_jobs):
            # Launch pipeline structure composition
            with self.timer.launch_composing():
                self.log.message('Pipeline composition started.')
                self.was_optimised = False
                best_pipelines = gp_composer.compose_pipeline(data=train_data)
                best_pipeline_candidates = gp_composer.best_models
                self.was_optimised = True
        else:
            # Use initial pipeline as final solution
            self.log.message(f'Timeout is too small for composing and is skipped '
                             f'because fit_time is {self.timer.assumption_fit_spend_time.total_seconds()} sec.')
            best_pipelines = fitted_assumption
            best_pipeline_candidates = [fitted_assumption]

        for pipeline in best_pipeline_candidates:
            pipeline.log = self.log
        best_pipeline = best_pipelines[0] if isinstance(best_pipelines, Sequence) else best_pipelines
        return best_pipeline, best_pipeline_candidates, gp_composer

    def tune_final_pipeline(self, task: Task,
                            train_data: InputData,
                            metric_function: Optional[MetricType],
                            composer_requirements: PipelineComposerRequirements,
                            pipeline_gp_composed: Pipeline,
                            ) -> Pipeline:
        """ Launch tuning procedure for obtained pipeline by composer """
        timeout_for_tuning = abs(self.timer.determine_resources_for_tuning()) / 60
        tuner = TunerBuilder(task) \
            .with_tuner(PipelineTuner) \
            .with_metric(metric_function) \
            .with_iterations(DEFAULT_TUNING_ITERATIONS_NUMBER) \
            .with_timeout(datetime.timedelta(minutes=timeout_for_tuning)) \
            .with_eval_time_constraint(composer_requirements.max_pipeline_fit_time) \
            .with_requirements(composer_requirements) \
            .build(train_data)

        if self.timer.have_time_for_tuning():
            # Tune all nodes in the pipeline
            with self.timer.launch_tuning():
                self.was_tuned = False
                self.log.message(f'Hyperparameters tuning started with {round(timeout_for_tuning)} min. timeout')
                tuned_pipeline = tuner.tune(pipeline_gp_composed)
                self.was_tuned = True
                self.log.message('Hyperparameters tuning finished')
        else:
            self.log.message(f'Time for pipeline composing was {str(self.timer.composing_spend_time)}.\n'
                             f'The remaining {max(0, round(timeout_for_tuning, 1))} seconds are not enough '
                             f'to tune the hyperparameters.')
            self.log.message('Composed pipeline returned without tuning.')
            tuned_pipeline = pipeline_gp_composed

        return tuned_pipeline


def _divide_parameters(common_dict: dict) -> List[dict]:
    """ Divide common dictionary into dictionary with parameters for API, Composer and Tuner

    :param common_dict: dictionary with parameters for all AutoML modules
    """
    api_params_dict = dict(train_data=None, task=Task, timeout=5, n_jobs=1, show_progress=True, logger=None)

    composer_params_dict = dict(max_depth=None, max_arity=None, pop_size=None, num_of_generations=None,
                                keep_n_best=None, available_operations=None, metric=None,
                                validation_blocks=None, cv_folds=None, genetic_scheme=None, history_folder=None,
                                early_stopping_iterations=None, early_stopping_timeout=None, optimizer=None,
                                optimizer_external_params=None, collect_intermediate_metric=False,
                                max_pipeline_fit_time=None, initial_assumption=None, preset='auto',
                                use_pipelines_cache=True, use_preprocessing_cache=True, cache_folder=None,
                                keep_history=True, history_dir=None,)

    tuner_params_dict = dict(with_tuning=False)

    dict_list = [api_params_dict, composer_params_dict, tuner_params_dict]
    for k, v in common_dict.items():
        is_unknown_key = True
        for i, dct in enumerate(dict_list):
            if k in dict_list[i]:
                dict_list[i][k] = v
                is_unknown_key = False
        if is_unknown_key:
            raise KeyError(f"Invalid key parameter {k}")

    return dict_list



import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc, mean_squared_error, mean_absolute_error

from examples.advanced.time_series_forecasting.composing_pipelines import visualise
from examples.simple.pipeline_import_export import create_correct_path
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from test.unit.tasks.test_classification import get_iris_data
from test.unit.tasks.test_forecasting import get_ts_data
from test.unit.tasks.test_regression import get_synthetic_regression_data, get_rmse_value


def pipeline_h2o_class() -> Pipeline:
    node = PrimaryNode('h2o_class')
    pipeline = Pipeline(node)
    return pipeline


def pipeline_h2o_regr() -> Pipeline:
    node = PrimaryNode('h2o_regr')
    pipeline = Pipeline(node)
    return pipeline


def pipeline_h2o_ts(window_size: int = 20):
    node_lagged = PrimaryNode('lagged')
    node_lagged.parameters = {'window_size': window_size}
    node_root = SecondaryNode('h2o_regr', nodes_from=[node_lagged])

    pipeline = Pipeline(node_root)

    return pipeline


def export_h2o(pipeline, pipeline_path, test_data):
    # Export it
    pipeline.save(path=pipeline_path)

    # Import pipeline
    json_path_load = create_correct_path(pipeline_path)
    new_pipeline = Pipeline.from_serialized(json_path_load)

    results = new_pipeline.predict(input_data=test_data, output_mode="full_probs")
    prediction_after_export = results.predict[:, 0]
    print(f'After export {prediction_after_export[:4]}')


def h2o_classification_pipeline_evaluation():
    pipeline_path = "h2o_class"
    data = get_iris_data()
    pipeline = pipeline_h2o_class()
    train_data, test_data = train_test_data_setup(data, shuffle_flag=True)

    pipeline.fit(input_data=train_data)
    results = pipeline.predict(input_data=test_data, output_mode="full_probs")
    prediction_before_export = results.predict[:, 0]
    print(f'Before export {prediction_before_export[:4]}')
    roc_auc_on_test = roc_auc(y_true=test_data.target,
                              y_score=results.predict,
                              multi_class='ovo',
                              average='macro')
    #  H2o has troubles with serialization for now
    #  export_h2o(pipeline, pipeline_path, test_data)
    print(f"roc auc: {roc_auc_on_test}")


def h2o_regression_pipeline_evaluation():
    data = get_synthetic_regression_data()

    pipeline = pipeline_h2o_regr()
    train_data, test_data = train_test_data_setup(data)

    pipeline.fit(input_data=train_data)
    results = pipeline.predict(input_data=test_data)
    _, rmse_on_test = get_rmse_value(pipeline, train_data, test_data)
    print(f"RMSE {rmse_on_test}")


def h2o_ts_pipeline_evaluation(visualization=False):
    train_data, test_data = get_ts_data(n_steps=500, forecast_length=3)

    pipeline = pipeline_h2o_ts()
    pipeline.fit(input_data=train_data)
    test_pred = pipeline.predict(input_data=test_data)
    # Calculate metric
    test_pred = np.ravel(np.array(test_pred.predict))
    test_target = np.ravel(np.array(test_data.target))

    plot_info = [{'idx': train_data.idx,
                  'series': train_data.features,
                  'label': 'Actual time series'},
                 {'idx': test_data.idx,
                  'series': test_target,
                  'label': 'Test part'}]
    metrics_info = {}
    rmse = mean_squared_error(test_target, test_pred, squared=False)
    mae = mean_absolute_error(test_target, test_pred)

    metrics_info['Metrics'] = {'RMSE': round(rmse, 3),
                               'MAE': round(mae, 3)}
    if visualization:
        visualise(plot_info)
    print(metrics_info)


if __name__ == '__main__':
    with OperationTypesRepository.init_automl_repository() as _:
        h2o_classification_pipeline_evaluation()
    with OperationTypesRepository.init_automl_repository() as _:
        h2o_regression_pipeline_evaluation()
    with OperationTypesRepository.init_automl_repository() as _:
        h2o_ts_pipeline_evaluation(visualization=True)



from typing import List, Type, Union

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.node import Node
from fedot.core.pipelines.pipeline import Pipeline
from fedot.sensitivity.node_sa_approaches import NodeAnalyzeApproach
from fedot.sensitivity.nodes_sensitivity import NodesAnalysis
from fedot.sensitivity.operations_hp_sensitivity.multi_operations_sensitivity import MultiOperationsHPAnalyze
from fedot.sensitivity.pipeline_sensitivity import PipelineAnalysis
from fedot.sensitivity.sa_requirements import SensitivityAnalysisRequirements


class PipelineSensitivityAnalysis:
    """This class works as facade and allows to apply all kind of approaches
    to whole pipeline and separate nodes together

    Args:
        pipeline: pipeline object to analyze
        train_data: data used for Pipeline training
        test_data: data used for Pipeline validation
        approaches: methods applied to pipeline. Default: ``None``
        nodes_to_analyze: nodes to analyze. Default: all nodes
        requirements: extra requirements to define specific details for different approaches.
            See :class:`SensitivityAnalysisRequirements` class documentation
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity/``
    """

    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 approaches: List[Union[Type[NodeAnalyzeApproach],
                                        Type[MultiOperationsHPAnalyze]]] = None,
                 nodes_to_analyze: List[Node] = None,
                 requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None):

        self.log = default_log(self)

        if approaches:
            nodes_analyze_approaches = [approach for approach in approaches
                                        if issubclass(approach, NodeAnalyzeApproach)]
            pipeline_analyze_approaches = [approach for approach in approaches
                                           if not issubclass(approach, NodeAnalyzeApproach)]
        else:
            self.log.info('Approaches for analysis are not given, thus will be set to defaults.')
            nodes_analyze_approaches = None
            pipeline_analyze_approaches = None

        self._nodes_analyze = NodesAnalysis(pipeline=pipeline,
                                            train_data=train_data,
                                            test_data=test_data,
                                            approaches=nodes_analyze_approaches,
                                            requirements=requirements,
                                            nodes_to_analyze=nodes_to_analyze,
                                            path_to_save=path_to_save, log=log)

        self._pipeline_analyze = PipelineAnalysis(pipeline=pipeline,
                                                  train_data=train_data,
                                                  test_data=test_data,
                                                  approaches=pipeline_analyze_approaches,
                                                  requirements=requirements,
                                                  path_to_save=path_to_save,
                                                  log=log)

    def analyze(self):
        """Applies defined sensitivity analysis approaches
        """

        if self._nodes_analyze:
            self._nodes_analyze.analyze()

        if self._pipeline_analyze:
            self._pipeline_analyze.analyze()



from copy import deepcopy
from os.path import join
from typing import List, Optional, Union

import numpy as np

from fedot.core.composer.metrics import MSE
from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.operations_hp_sensitivity.problem import MultiOperationsProblem, Problem
from fedot.sensitivity.operations_hp_sensitivity.sa_and_sample_methods import (
    analyze_method_by_name,
    sample_method_by_name
)
from fedot.sensitivity.sa_requirements import HyperparamsAnalysisMetaParams, SensitivityAnalysisRequirements


class MultiOperationsHPAnalyze:
    """Provides with analysis of all the :class:`Pipeline`'s operations hyperparameters
    using sample and analyze methods from ``SALib``

    Args:
        pipeline: :obj:`Pipeline` object to analyze
        train_data: data used for :obj:`Pipeline` training
        test_data: data used for :obj:`Pipeline` validation
        requirements: extra requirements to define specific details for different approaches.
            See :class:`SensitivityAnalysisRequirements` class documentation.
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity/``
    """

    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None):
        self._pipeline = pipeline
        self._train_data = train_data
        self._test_data = test_data
        self.problem: Optional[Problem] = None
        requirements = SensitivityAnalysisRequirements() if requirements is None else requirements
        self.requirements: HyperparamsAnalysisMetaParams = requirements.hp_analysis_meta
        self.analyze_method = analyze_method_by_name.get(self.requirements.analyze_method)
        self.sample_method = sample_method_by_name.get(self.requirements.sample_method)

        self.operation_types = None
        self.path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'pipeline_sensitivity') \
                if path_to_save is None else path_to_save
        self.log = default_log(self)

    def analyze(self) -> dict:
        """Analyze all the hyperparameters af all :obj:`Pipeline` operations using ``SA`` methods.\n
        Default: Sobol method with Saltelli sample algorithm

        Returns:
            dict: ``Main`` and total ``Sobol`` indices for every parameter per node
        """

        if not self._pipeline.is_fitted:
            self._pipeline.fit(self._train_data)

        # create problem
        self.operation_types = [node.operation.operation_type for node in self._pipeline.nodes]
        self.problem = MultiOperationsProblem(self.operation_types)

        # sample
        self.log.info('Making hyperparameters samples')
        samples = self.sample(self.requirements.sample_size)
        response_matrix = self._get_response_matrix(samples)

        self.log.info('Start hyperparameters sensitivity analysis')
        indices = self.analyze_method(self.problem.dictionary, samples, response_matrix)
        converted_to_json_indices = self.convert_results_to_json(problem=self.problem,
                                                                 si=indices)

        self.log.info('Finish hyperparameters sensitivity analysis')

        return converted_to_json_indices

    def sample(self, *args) -> Union[List[Pipeline], Pipeline]:
        """Makes hyperparameters samples

        Args:
            args: i.e. ``sample_size``
        Returns:
            List[Pipeline]: new sampled hyperparameters
        """

        sample_size = args[0]
        samples = self.sample_method(self.problem.dictionary, num_of_samples=sample_size)
        converted_samples: List[List[dict]] = self.problem.convert_sample_to_dict(samples)
        sampled_pipelines: List[Pipeline] = self._apply_params_to_node(params=converted_samples)

        return sampled_pipelines

    def _apply_params_to_node(self, params: List[List[dict]]) -> List[Pipeline]:
        sampled_pipelines: List[Pipeline] = list()
        for sample in params:
            copied_pipeline = deepcopy(self._pipeline)
            for node_id, params_per_node in enumerate(sample):
                copied_pipeline.nodes[node_id].parameters = params_per_node
                sampled_pipelines.append(copied_pipeline)

        return sampled_pipelines

    def _get_response_matrix(self, samples: List[Pipeline]):
        operation_response_matrix = []
        for sampled_pipeline in samples:
            sampled_pipeline.fit(self._train_data)
            prediction = sampled_pipeline.predict(self._test_data)
            mse_metric = MSE().metric(reference=self._test_data,
                                      predicted=prediction)
            operation_response_matrix.append(mse_metric)

        return np.array(operation_response_matrix)

    @staticmethod
    def convert_results_to_json(problem: MultiOperationsProblem, si: dict) -> dict:
        sobol_indices = []
        for index, param_name in enumerate(problem.names):
            var_indices = {f"{param_name}": {
                'S1': list(si['S1'])[index],
                'S1_conf': list(si['S1_conf'])[index],
                'ST': list(si['ST'])[index],
                'ST_conf': list(si['ST_conf'])[index],
            }}
            sobol_indices.append(var_indices)

        indices_per_operation = dict()
        border_index = 0
        for operation_name, params_names in problem.names_per_node.items():
            if params_names is not None:
                indices_per_operation[operation_name] = \
                    sobol_indices[border_index:border_index + len(params_names)]
                border_index += len(params_names)
            else:
                indices_per_operation[operation_name] = 'None'

        data = {
            'problem': problem.dictionary,
            'sobol_indices_per_operation': indices_per_operation
        }

        return data



from copy import copy
from typing import Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import ts_to_table
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class PolyfitImplementation(ModelImplementation):
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.min_degree = 1
        self.max_degree = 5
        self.default_degree = 3
        self._correct_degree()
        self.coefs = None

    def _correct_degree(self):
        degree = self.params.get('degree')
        if not degree or not self.min_degree <= degree <= self.max_degree:
            # default value
            self.log.debug(f"Change invalid parameter degree ({degree}) on default value (3)")
            degree = self.default_degree
            self.params.update(degree=degree)

    @property
    def degree(self):
        return int(self.params.get('degree'))

    def fit(self, input_data):
        f_x = input_data.idx
        f_y = input_data.features
        self.coefs = np.polyfit(f_x, f_y, deg=self.degree)

        return self.coefs

    def predict(self, input_data):
        f_x = input_data.idx
        f_y = np.polyval(self.coefs, f_x)
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        old_idx = input_data.idx

        start_id = old_idx[-1] - forecast_length + 1
        end_id = old_idx[-1]
        predict = f_y
        predict = np.array(predict).reshape(1, -1)
        new_idx = np.arange(start_id, end_id + 1)

        input_data.idx = new_idx

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        f_x = input_data.idx
        f_y = np.polyval(self.coefs, f_x)
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        old_idx = input_data.idx
        target = input_data.target

        _, predict = ts_to_table(idx=old_idx,
                                 time_series=f_y,
                                 window_size=forecast_length)
        new_idx, target_columns = ts_to_table(idx=old_idx,
                                              time_series=target,
                                              window_size=forecast_length)

        # Update idx and target
        input_data.idx = new_idx
        input_data.target = target_columns
        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data



import os
import random
from datetime import datetime

import numpy as np

from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import fedot_project_root
from fedot.remote.infrastructure.clients.test_client import TestClient
from fedot.remote.remote_evaluator import RemoteEvaluator, RemoteTaskParams

random.seed(1)
np.random.seed(1)

num_parallel = 3  # NUMBER OF PARALLEL TASKS

# WARNING - THIS SCRIPT CAN BE EVALUATED ONLY WITH THE ACCESS TO DATAMALL SYSTEM

# LOCAL RUN
folder = os.path.join(fedot_project_root(), 'cases', 'data', 'scoring')
path = os.path.join(folder, 'scoring_train.csv')

start = datetime.now()
data = InputData.from_csv(path)
data.subset_indices([1, 2, 3, 4, 5, 20])
pipeline = Pipeline(PrimaryNode('rf'))
pipeline.fit_from_scratch(data)
end = datetime.now()

print('LOCAL EXECUTION TIME', end - start)

# REMOTE RUN

connect_params = {}
exec_params = {
    'container_input_path': folder,
    'container_output_path': os.path.join(folder, 'remote'),
    'container_config_path': ".",
    'container_image': "test",
    'timeout': 1
}

remote_task_params = RemoteTaskParams(
    mode='remote',
    dataset_name='scoring_train',
    task_type='Task(TaskTypesEnum.classification)',
    max_parallel=num_parallel
)

client = TestClient(connect_params, exec_params, output_path=os.path.join(folder, 'remote'))

evaluator = RemoteEvaluator()
evaluator.init(
    client=client,
    remote_task_params=remote_task_params
)

pipelines = [Pipeline(PrimaryNode('rf'))] * num_parallel
fitted_pipelines = evaluator.compute_graphs(pipelines)

[print(p.is_fitted) for p in fitted_pipelines]



from typing import Dict, Union
from typing import Optional

import numpy as np

from fedot.api.api_utils.data_definition import data_strategy_selector, FeaturesType, TargetType
from fedot.core.data.data import InputData, OutputData, data_type_is_table
from fedot.core.data.data_preprocessing import convert_into_column
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.ts_wrappers import in_sample_ts_forecast, convert_forecast_to_output
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.preprocessing.preprocessing import DataPreprocessor


class ApiDataProcessor:
    """
    Class for selecting optimal data processing strategies based on type of data.
    Available data sources are:
        * numpy array
        * pandas DataFrame
        * string (path to csv file)
        * InputData (FEDOT dataclass)

    Data preprocessing such a class performing also
    """

    def __init__(self, task: Task):
        self.task = task
        self.preprocessor = DataPreprocessor()

        # Dictionary with recommendations (e.g. 'cut' for cutting dataset, 'label_encode'
        # to encode features using label encoder). Parameters for transformation provided also
        self.recommendations = {'cut': self.preprocessor.cut_dataset,
                                'label_encoded': self.preprocessor.label_encoding_for_fit}

    def define_data(self,
                    features: FeaturesType,
                    target: Optional[TargetType] = None,
                    is_predict=False):
        """ Prepare data for FEDOT pipeline composing.
        Obligatory preprocessing steps are applying also. If features is dictionary
        there is a need to process MultiModalData
        """
        try:
            # TODO remove workaround
            idx = None
            if isinstance(features, dict) and 'idx' in features:
                idx = features['idx']
                del features['idx']
            data = data_strategy_selector(features=features,
                                          target=target,
                                          ml_task=self.task,
                                          is_predict=is_predict)
            if isinstance(data, dict) and idx is not None:
                for key in data:
                    data[key].idx = idx
        except Exception as ex:
            raise ValueError('Please specify the "features" as path to as path to csv file/'
                             'Numpy array/Pandas DataFrame/FEDOT InputData/dict for multimodal data, '
                             f'Exception: {ex}')

        # Perform obligatory steps of data preprocessing
        if is_predict:
            data = self.preprocessor.obligatory_prepare_for_predict(data)
        else:
            data = self.preprocessor.obligatory_prepare_for_fit(data)
        return data

    def define_predictions(self, current_pipeline: Pipeline, test_data: Union[InputData, MultiModalData],
                           in_sample: bool = False, validation_blocks: int = None) -> OutputData:
        """ Prepare predictions """
        if self.task.task_type == TaskTypesEnum.classification:
            # Prediction should be converted into source labels
            output_prediction = current_pipeline.predict(test_data, output_mode='labels')
        elif self.task.task_type == TaskTypesEnum.ts_forecasting:
            if in_sample:
                forecast_length = test_data.task.task_params.forecast_length
                validation_blocks = validation_blocks or 1
                horizon = forecast_length * validation_blocks
                forecast = in_sample_ts_forecast(current_pipeline, test_data, horizon)
                idx = test_data.idx[-horizon:]
                prediction = convert_forecast_to_output(test_data, forecast, idx=idx)
            else:
                prediction = current_pipeline.predict(test_data)
                # Convert forecast into one-dimensional array
                forecast = np.ravel(np.array(prediction.predict))
                prediction.predict = forecast
            output_prediction = prediction
        else:
            output_prediction = current_pipeline.predict(test_data)

        return output_prediction

    def correct_predictions(self, real: InputData, prediction: OutputData):
        """ Change shape for models predictions if its necessary. Apply """
        if self.task == TaskTypesEnum.ts_forecasting:
            real.target = real.target[~np.isnan(prediction.predict)]
            prediction.predict = prediction.predict[~np.isnan(prediction.predict)]

        if data_type_is_table(prediction):
            # Check dimensions for real and predicted values
            if len(real.target.shape) != len(prediction.predict.shape):
                prediction.predict = convert_into_column(prediction.predict)
                real.target = convert_into_column(np.array(real.target))

    def accept_and_apply_recommendations(self, input_data: Union[InputData, MultiModalData], recommendations: Dict):
        """
        Accepts recommendations for preprocessing from DataAnalyser

        :param input_data - data for preprocessing
        :param recommendations - dict with recommendations
        """
        if isinstance(input_data, MultiModalData):
            for data_source_name, values in input_data.items():
                self.accept_and_apply_recommendations(input_data[data_source_name], recommendations[data_source_name])
        else:
            for name in recommendations:
                rec = recommendations[name]
                # Apply desired preprocessing function
                self.recommendations[name](input_data, *rec.values())



import numpy as np
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.model_selection._split import _BaseKFold
from typing import Type

from fedot.core.data.data import InputData
from fedot.core.repository.tasks import TaskTypesEnum


class DataObjectiveAdvisor:
    def __init__(self, threshold: float = 0.5):
        """
        Advisor for DataObjectiveBuilder for choice some parameters based on input_data

        :param threshold: threshold level for difference between uniform probabilities and real probabilities
        """
        self.threshold = threshold

    def propose_kfold(self, input_data: InputData) -> Type[_BaseKFold]:
        """
        Method to choose he most suitable strategy for making folds

        :param input_data: data to analyse
        """
        if input_data.task.task_type is TaskTypesEnum.classification and self.check_imbalance(input_data):
            return StratifiedKFold
        else:
            return KFold

    def check_imbalance(self, input_data: InputData) -> bool:
        """
        Checks data for imbalance - if probability of any class lower than uniform probability in threshold times it
        returns true
        :param input_data: data to analyse

        """
        _, counts = np.unique(input_data.target, return_counts=True)
        probabilities = counts / input_data.target.shape[0]
        uniform_probability = 1 / input_data.num_classes
        return np.any(np.abs(uniform_probability - probabilities) / uniform_probability > self.threshold)



import dataclasses
from dataclasses import dataclass
from numbers import Number
from typing import Sequence

from fedot.core.optimisers.composer_requirements import ComposerRequirements


@dataclass
class PipelineComposerRequirements(ComposerRequirements):
    """Defines restrictions and requirements for composition of final graphs.

    Restrictions on final graphs:
    :param start_depth: start value of adaptive tree depth
    :param max_depth: max depth of the resulting pipeline
    :param min_arity: min number of parents for node
    :param max_arity: max number of parents for node

    :param primary: operation types for :class:`~fedot.core.pipelines.node.PrimaryNode`s
    :param secondary: operation types for :class:`~fedot.core.pipelines.node.SecondaryNode`s

    """

    start_depth: int = 3
    max_depth: int = 3
    min_arity: int = 2
    max_arity: int = 2

    primary: Sequence[str] = tuple()
    secondary: Sequence[str] = tuple()

    def __post_init__(self):
        super().__post_init__()
        excluded_fields = ['n_jobs']
        for field_name, field_value in dataclasses.asdict(self).items():
            if field_name in excluded_fields:
                continue
            if isinstance(field_value, Number) and field_value < 0:
                raise ValueError(f'Value of {field_name} must be non-negative')



from typing import Tuple

from fedot.core.data.data import InputData, OutputData

from test.unit.models.test_split_train_test import get_synthetic_input_data
from fedot.core.operations.evaluation.gpu.common import CuMLEvaluationStrategy
from fedot.core.operations.evaluation.gpu.classification import CuMLClassificationStrategy
from cuml.svm import SVC


def get_synthetic_data() -> Tuple[InputData, InputData]:
    train_data = get_synthetic_input_data(10000)
    test_data = get_synthetic_input_data(1000)

    return train_data, test_data


def test_gpu_evaluation_strategy_fit():
    train_data, _ = get_synthetic_data()
    strategy = CuMLEvaluationStrategy(operation_type='svc',
                                      params=dict(kernel='rbf', C=10, gamma=1, cache_size=2000, probability=True))

    operation = strategy.fit(train_data)

    assert isinstance(operation, SVC)


def test_gpu_evaluation_strategy_predict():
    train_data, test_data = get_synthetic_data()
    strategy = CuMLClassificationStrategy(operation_type='svc',
                                          params=dict(kernel='rbf', C=10, gamma=1, cache_size=2000, probability=True))

    operation = strategy.fit(train_data)
    prediction = strategy.predict(operation, test_data)

    assert isinstance(prediction, OutputData)
